"""
Copyright (c) 2016, Maarten Everts
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/gabi
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""

from binascii import hexlify
from os import urandom

from cryptography.hazmat.primitives.asymmetric.rsa import _modinv  # type:ignore

from .. import secure_randint
from ...primitives.attestation import sha256_as_int
from ...primitives.cryptography_wrapper import generate_safe_prime, is_prime
from ...primitives.value import FP2Value
from .....util import byte2int, int2byte

DefaultEpochLength = 432000


class BaseParameters(object):

    def __init__(self, LePrime, Lh, Lm, Ln, Lstatzk):
        self.LePrime = LePrime
        self.Lh = Lh
        self.Lm = Lm
        self.Ln = Ln
        self.Lstatzk = Lstatzk

        self.Lv = Ln + 2 * Lstatzk + Lh + Lm + 4
        self.Le = Lstatzk + Lh + Lm + 5
        self.LeCommit = LePrime + Lstatzk + Lh
        self.LmCommit = Lm + Lstatzk + Lh
        self.LRA = Ln + Lstatzk
        self.LsCommit = Lm + Lstatzk + Lh + 1
        self.LvCommit = self.Lv + Lstatzk + Lh
        self.LvPrime = Ln + Lstatzk
        self.LvPrimeCommit = Ln + 2 * Lstatzk + Lh


DefaultSystemParameters = {
    1024: BaseParameters(120, 256, 256, 1024, 80),
    2048: BaseParameters(120, 256, 256, 2048, 128),
    4096: BaseParameters(120, 256, 512, 4096, 128)
}


class PrivateKey(object):

    def __init__(self, p, q, counter, expiryDate):
        self.Counter = counter
        self.ExpiryData = expiryDate
        self.P = p
        self.Q = q
        self.PPrime = (p - 1) >> 1
        self.QPrime = (q - 1) >> 1


class PublicKey(object):

    def __init__(self, N, Z, S, R, counter, expiryDate, param=None):
        self.Counter = counter
        self.ExpiryDate = expiryDate
        self.N = N
        self.Z = Z
        self.S = S
        self.R = R
        self.EpochLength = DefaultEpochLength
        self.Params = param or DefaultSystemParameters[N.bit_length()]
        self.Issuer = "-"


def findMatch(safeprimes, param, p):
    for q in safeprimes:
        if (p * q).bit_length() == param.Ln and (p % 8) != 0 and (q % 8) != 0:
            return q
    return None


def generateSafePrimePair(param):
    primeSize = param.Ln // 2
    safeprimes = []
    while True:
        p = generate_safe_prime(primeSize)
        pPrime = p >> 1
        pPrimeMod8 = pPrime % 8
        if pPrimeMod8 == 1:
            continue
        q = findMatch(safeprimes, param, p)
        if not safeprimes or not q:
            safeprimes.append(p)
            continue
        break
    return p, q


def legendreSymbol(a, p):
    j = 1
    n = a % p
    m = p
    while n != 0:
        t = 0
        while (n & 1) == 0:
            n >>= 1
            t += 1
        tmp = m % 8
        if (t & 1) == 1 and (tmp == 3 or tmp == 5):
            j = -j
        if (m % 4) == 3 or (n % 4) == 3:
            j = -j
        m = m % n
        n, m = m, n
    if m == 1:
        return j
    return 0


def GenerateKeyPair(param, numAttributes, counter, expiryDate):
    p, q = generateSafePrimePair(param)
    priv = PrivateKey(p, q, counter, expiryDate)

    N = p * q

    while True:
        S = secure_randint(param.Ln)
        if S > N:
            continue
        if legendreSymbol(S, p) == 1 and legendreSymbol(S, q) == 1:
            break

    primeSize = param.Ln // 2
    while True:
        x = secure_randint(primeSize)
        if x > 2 and x < N:
            break

    Z = FP2Value(N, S).intpow(x).a

    R = []

    for i in range(numAttributes):
        while True:
            x = secure_randint(primeSize)
            if x > 2 and x < N:
                break
        R.append(FP2Value(N, S).intpow(x).a)

    pubk = PublicKey(N, Z, S, R, counter, expiryDate, param=param)

    if not SignMessageBlock(priv, pubk, [1]).Verify(pubk, [1]):
        return GenerateKeyPair(param, numAttributes, counter, expiryDate)
    else:
        return priv, pubk


class CLSignature(object):

    def __init__(self, A, E, V, KeyshareP=None):
        self.A = A
        self.E = E
        self.V = V
        self.KeyshareP = KeyshareP

    def Verify(self, pk, ms):
        start = 1 << (pk.Params.Le - 1)
        end = 1 << (pk.Params.LePrime - 1)
        end = end + start

        if self.E < start or self.E > end:
            return False

        Ae = FP2Value(pk.N, self.A).intpow(self.E).a
        R = RepresentToPublicKey(pk, ms)

        if self.KeyshareP is not None:
            R = R * self.KeyshareP

        Sv = FP2Value(pk.N, pk.S).intpow(self.V).a
        Q = Ae * R
        Q = (Q * Sv) % pk.N

        return pk.Z == Q

    def Randomize(self, pk):
        r = secure_randint(pk.Params.LRA)
        APrime = (FP2Value(pk.N, self.A) * FP2Value(pk.N, pk.S).intpow(r)).a
        t = self.E * r
        VPrime = self.V - t
        return CLSignature(APrime, self.E, VPrime, None)


def representToBases(bases, exps, modulus, maxMessageLength):
    r = 1
    for i in range(len(exps)):
        exp = exps[i]
        if exp.bit_length() > maxMessageLength:
            exp = sha256_as_int(str(exp))
        tmp = FP2Value(modulus, bases[i]).intpow(exp).a
        r = (r * tmp) % modulus
    return r


def RepresentToPublicKey(pk, exps):
    return representToBases(pk.R, exps, pk.N, pk.Params.Lm)


smallPrimes = [3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53]
smallPrimesProduct = 16294579238595022365


def randomPrimeInRange(start, length):
    b = length % 8
    if b == 0:
        b = 8
    startVal = 1 << start
    endVal = (1 << length) + startVal

    while True:
        bytez = urandom((length + 7) // 8)
        bytez = int2byte(byte2int(bytez[0:1]) & (1 << b) - 1) + bytez[1:]
        bytez = bytez[:-1] + int2byte(byte2int(bytez[-1:]) | 1)
        offset = int(hexlify(bytez), 16)
        p = startVal + offset
        bigMod = p % smallPrimesProduct
        mod = bigMod

        delta = 0
        while delta < (1 << 20):
            m = mod + delta
            if any(m % prime == 0 and (start > 6 or m != prime) for prime in smallPrimes):
                break
            if delta > 0:
                bigMod = delta
                p += bigMod
            delta += 2

        if is_prime(p) and p < endVal:
            return p


def signMessageBlockAndCommitment(sk, pk, U, ms):
    R = RepresentToPublicKey(pk, ms)
    vTilde = secure_randint(pk.Params.Lv)
    twoLv = 1 << (pk.Params.Lv - 1)
    v = twoLv + vTilde

    numerator = FP2Value(pk.N, pk.S).intpow(v).a
    numerator = (numerator * R * U) % pk.N

    invNumerator = _modinv(numerator, pk.N)
    Q = (pk.Z * invNumerator) % pk.N

    e = randomPrimeInRange(pk.Params.Le - 1, pk.Params.LePrime - 1)

    order = sk.PPrime * sk.QPrime
    d = _modinv(e, order)
    A = FP2Value(pk.N, Q).intpow(d).a

    return CLSignature(A, e, v, None)


def SignMessageBlock(sk, pk, ms):
    return signMessageBlockAndCommitment(sk, pk, 1, ms)



import time

from .network_stats import NetworkStat
from ...messaging.interfaces.endpoint import EndpointListener


class StatisticsEndpoint(EndpointListener):
    """
    This class is responsible for keeping stats regarding community.
    The stats are basically of the messages that the community can handle.

    This endpoint acts both as wrapper for IPv8 Endpoint, and EndpointListener.
    It inherits from EndpointListener directly and implements on_packet(self, packet)
    to measure statistics about received data. But, all the functionality of Endpoint
    itself is delegated to the existing IPv8 UDPEndpoint.
    """

    IDS_INTRODUCTION = [245, 246]
    IDS_PUNCTURE = [249, 250]
    IDS_DEPRECATED = [235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 247, 248, 251, 252, 253, 254, 255]

    def __init__(self, endpoint):
        EndpointListener.__init__(self, endpoint)
        self.endpoint = endpoint
        self.endpoint.add_listener(self)
        self.statistics = {}

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return object.__getattribute__(self.endpoint, item)

    # Endpoint methods
    def send(self, socket_address, packet):
        self.endpoint.send(socket_address, packet)

        prefix = packet[:22]
        if prefix not in list(self.statistics.keys()) or len(packet) < 22:
            return

        self.add_sent_stat(prefix, packet[22], len(packet))

    def enable_community_statistics(self, community_prefix, enabled):
        if community_prefix not in self.statistics and enabled:
            self.statistics[community_prefix] = {}
        elif community_prefix in self.statistics and not enabled:
            self.statistics.pop(community_prefix)

    # EndpointListener methods
    def on_packet(self, packet):
        _, data = packet

        prefix = data[:22]
        if prefix not in list(self.statistics.keys()) or len(data) < 22:
            return

        message_id = data[22]
        self.add_received_stat(prefix, message_id, len(data))

    # Statistics methods
    def add_sent_stat(self, prefix, identifier, num_bytes, timestamp=None):
        if prefix not in self.statistics:
            self.statistics[prefix] = {}
        if identifier not in self.statistics[prefix]:
            self.statistics[prefix][identifier] = NetworkStat(identifier)
        self.statistics[prefix][identifier].add_sent_stat(timestamp if timestamp else time.time(), num_bytes)

    def add_received_stat(self, prefix, identifier, num_bytes, timestamp=None):
        if prefix not in self.statistics:
            self.statistics[prefix] = {}
        if identifier not in self.statistics[prefix]:
            self.statistics[prefix][identifier] = NetworkStat(identifier)
        self.statistics[prefix][identifier].add_received_stat(timestamp if timestamp else time.time(), num_bytes)

    def get_statistics(self, prefix):
        if prefix in self.statistics:
            return self.statistics[prefix]
        return {}

    def get_aggregate_statistics(self, prefix):
        def min_positive(a, b):
            return a if not b else b if b < a or not a else a

        aggregate_stats = {
            "num_up": 0,
            "num_down": 0,
            "bytes_up": 0,
            "bytes_down": 0,
            "diff_time": 0
        }

        if prefix in self.statistics:
            first_ts = last_ts = 0
            for _, network_stat in self.statistics[prefix].items():
                aggregate_stats['num_up'] += network_stat.num_up
                aggregate_stats['num_down'] += network_stat.num_down
                aggregate_stats['bytes_up'] += network_stat.bytes_up
                aggregate_stats['bytes_down'] += network_stat.bytes_down

                first_ts = min_positive(first_ts, min_positive(network_stat.first_measured_up,
                                                               network_stat.first_measured_down))
                last_ts = max(last_ts, max(network_stat.last_measured_up, network_stat.last_measured_down))

            aggregate_stats["diff_time"] = last_ts - first_ts
        return aggregate_stats

    def get_message_sent(self, prefix, include_introduction=False, include_puncture=False, include_deprecated=False):
        num_sent = 0
        if prefix in self.statistics:
            for identifier in self.statistics[prefix]:
                if not self.is_excluded(identifier, include_introduction, include_puncture, include_deprecated):
                    num_sent += self.statistics[prefix][identifier].num_up
        return num_sent

    def get_message_received(self, prefix, include_introduction=False, include_puncture=False,
                             include_deprecated=False):
        num_received = 0
        if prefix in self.statistics:
            for identifier in self.statistics[prefix]:
                if not self.is_excluded(identifier, include_introduction, include_puncture, include_deprecated):
                    num_received += self.statistics[prefix][identifier].num_down
        return num_received

    def get_bytes_sent(self, prefix, include_introduction=False, include_puncture=False, include_deprecated=False):
        bytes_sent = 0
        if prefix in self.statistics:
            for identifier in self.statistics[prefix]:
                if not self.is_excluded(identifier, include_introduction, include_puncture, include_deprecated):
                    bytes_sent += self.statistics[prefix][identifier].bytes_up
        return bytes_sent

    def get_bytes_received(self, prefix, include_introduction=False, include_puncture=False, include_deprecated=False):
        bytes_received = 0
        if prefix in self.statistics:
            for identifier in self.statistics[prefix]:
                if not self.is_excluded(identifier, include_introduction, include_puncture, include_deprecated):
                    bytes_received += self.statistics[prefix][identifier].bytes_down
        return bytes_received

    def is_excluded(self, identifier, include_introduction, include_puncture, include_deprecated):
        return (identifier in StatisticsEndpoint.IDS_DEPRECATED and not include_deprecated
                or identifier in StatisticsEndpoint.IDS_INTRODUCTION and not include_introduction
                or identifier in StatisticsEndpoint.IDS_PUNCTURE and not include_puncture)



import logging
from collections import OrderedDict, namedtuple
from threading import RLock
from typing import Dict, Iterable, List, Optional, Set

from ..messaging.serialization import default_serializer
from ..types import Address, Overlay, Peer


MID = bytes
PublicKeyMat = bytes
Service = bytes
ServiceSet = Set[Service]
WalkableAddress = namedtuple('WalkableAddress', ['introduced_by', 'services', 'new_style'])


class Network(object):

    def __init__(self):
        self._all_addresses: Dict[Address, WalkableAddress] = {}
        '''All known IP:port addresses, mapped to (introduction peer, services, new_style).'''

        self.verified_peers: Set[Peer] = set()
        '''All verified Peer objects (Peer.address must be in _all_addresses).'''

        self.verified_by_public_key_bin: Dict[PublicKeyMat, Peer] = {}
        '''Map of known public keys to Peer objects.'''

        self.graph_lock = RLock()
        '''Lock for all updates of the peer pool.'''

        self.blacklist: List[Address] = []
        '''Peers we should not add to the network (e.g., bootstrap peers), by address.'''

        self.blacklist_mids: List[MID] = []
        '''Peers we should not add to the network (e.g., bootstrap peers), by mid.'''

        self.services_per_peer: Dict[PublicKeyMat, ServiceSet] = {}
        '''Map of advertised services (set) per peer.'''

        self.service_overlays: Dict[Service, Overlay] = {}
        '''Map of service identifiers to local overlays.'''

        self.reverse_ip_cache_size = 500
        self.reverse_ip_lookup = OrderedDict()
        '''Cache of IP:port -> Peer. This is a cache rather than a normal dictionary (the addresses of a peer are
        temporal and can grow infinitely): we rotate out old information to avoid a memory leak.'''

        self.reverse_intro_cache_size = 500
        self.reverse_intro_lookup = OrderedDict()
        '''Map of Peer -> [IP:port], reversing the information from _all_addresses. This is a cache rather than a
        normal dictionary (the addresses of a peer are temporal and can grow infinitely): we rotate out old
        information to avoid a memory leak.'''

        self.reverse_service_cache_size = 500
        self.reverse_service_lookup = OrderedDict()
        '''Cache of service_id -> [Peer]. This is a cache rather than a normal dictionary (the services of a peer may
        be temporal and can grow infinitely): we rotate out old information to avoid a memory leak.'''

    def is_new_style(self, address: Address) -> bool:
        """
        Check if an address supports new-style introduction requests and responses.

        :param address: the address to check for.
        :returns: True iff the address is both known and known to support new-style introductions.
        """
        return (self._all_addresses.get(address) or WalkableAddress(b'', None, False)).new_style

    def discover_address(self,
                         peer: Peer,
                         address: Address,
                         service: Optional[Service] = None,
                         new_style: bool = False) -> None:
        """
        A peer has introduced us to another IP address.

        :param peer: the peer that performed the introduction.
        :param address: the introduced address.
        :param service: the service through which we discovered the peer.
        :param new_style: the introduced address uses new introduction logic.
        """
        if address in self.blacklist:
            self.add_verified_peer(peer)
            return

        with self.graph_lock:
            if ((address not in self._all_addresses)
                    or (self._all_addresses[address].introduced_by not in self.verified_by_public_key_bin)):
                # This is a new address, or our previous parent has been removed
                self._all_addresses[address] = WalkableAddress(peer.public_key.key_to_bin(), service, new_style)
                intro_cache = self.reverse_intro_lookup.get(peer, None)
                if intro_cache:
                    intro_cache.append(address)
                else:
                    self.reverse_intro_lookup[peer] = [address]

            self.add_verified_peer(peer)

    def discover_services(self, peer: Peer, services: Iterable) -> None:
        """
        A peer has advertised some services he can use.

        :param peer: the peer to update the services for.
        :param services: the list of services to register.
        """
        with self.graph_lock:
            key_material = peer.public_key.key_to_bin()
            if key_material not in self.services_per_peer:
                self.services_per_peer[key_material] = set(services)
            else:
                self.services_per_peer[key_material] |= set(services)
            for service in services:
                service_cache = self.reverse_service_lookup.get(service, [])
                # Ensure that the peer instance in the cache is the same one as in verified_peers.
                if peer in service_cache:
                    service_cache.remove(peer)
                service_cache.append(self.verified_by_public_key_bin.get(key_material, peer))
                self.reverse_service_lookup[service] = service_cache

    def add_verified_peer(self, peer: Peer) -> None:
        """
        The holepunching layer has a new peer for us.

        :param peer: the new peer.
        """
        if peer.mid in self.blacklist_mids:
            return
        with self.graph_lock:
            # This may just be an address update
            known = self.verified_by_public_key_bin.get(peer.public_key.key_to_bin(), None)
            if known:
                known.addresses.update(peer.addresses)
                return
            if any(address in self._all_addresses for address in peer.addresses.values()):
                if peer not in self.verified_peers:
                    # This should always happen, unless someone edits the verified_peers dict directly.
                    # This would be a programmer 'error', but we will allow it.
                    self.verified_peers.add(peer)
                    self.verified_by_public_key_bin[peer.public_key.key_to_bin()] = peer
            elif all(address not in self.blacklist for address in peer.addresses.values()):
                for address in peer.addresses.values():
                    if address not in self._all_addresses:
                        self._all_addresses[address] = WalkableAddress(b'', None, False)
                if peer not in self.verified_peers:
                    self.verified_peers.add(peer)
                    self.verified_by_public_key_bin[peer.public_key.key_to_bin()] = peer

    def register_service_provider(self, service_id: Service, overlay: Overlay) -> None:
        """
        Register an overlay to provide a certain service id.

        :param service_id: the name/id of the service.
        :param overlay: the actual service.
        """
        with self.graph_lock:
            self.service_overlays[service_id] = overlay

    def get_peers_for_service(self, service_id: Service) -> List[Peer]:
        """
        Get peers which support a certain service.

        :param service_id: the service name/id to fetch peers for.
        """
        out = []
        service_cache = self.reverse_service_lookup.pop(service_id, None)
        if service_cache is None:
            with self.graph_lock:
                for peer in self.verified_peers:
                    key_material = peer.public_key.key_to_bin()
                    if service_id in self.services_per_peer.get(key_material, []):
                        out.append(peer)
        else:
            out = [peer for peer in service_cache if
                   peer in self.verified_peers
                   and service_id in self.services_per_peer.get(peer.public_key.key_to_bin(), [])]
        self.reverse_service_lookup[service_id] = out
        if len(self.reverse_service_lookup) > self.reverse_service_cache_size:
            self.reverse_service_lookup.popitem(False)  # Pop the oldest cache entry
        return out

    def get_services_for_peer(self, peer):
        """
        Get the known services supported by a peer.

        :param peer: the peer to check services for.
        """
        with self.graph_lock:
            return self.services_per_peer.get(peer.public_key.key_to_bin(), set())

    def get_walkable_addresses(self,
                               service_id: Optional[Service] = None,
                               old_style: bool = False) -> List[Address]:
        """
        Get all addresses ready to be walked to.

        :param service_id: the service_id to filter on.
        :param old_style: only return addresses that are not new-style.
        """
        with self.graph_lock:
            known = self.get_peers_for_service(service_id) if service_id else self.verified_peers
            verified: List[Address] = []
            for peer in known:
                verified.extend(peer.addresses.values())
            out = list(set(self._all_addresses.keys()) - set(verified))
            if service_id:
                new_out = []
                for address in out:
                    intro_peer, service, new_style = self._all_addresses[address]
                    if old_style and new_style:
                        continue
                    services = self.services_per_peer.get(intro_peer, set([]))
                    if service:
                        services.add(service)
                    if service_id in services:
                        new_out.append(address)
                out = new_out
            return out

    def get_verified_by_address(self, address: Address) -> Peer:
        """
        Get a verified Peer by its IP address.

        If multiple Peers use the same IP address, this method returns only one of these peers.

        :param address: the (IP, port) tuple to search for
        :return: the Peer object for this address or None
        """
        with self.graph_lock:
            peer = self.reverse_ip_lookup.pop(address, None)
            if not peer:
                for p in self.verified_peers:
                    if address in p.addresses.values():
                        peer = p
                        self.reverse_ip_lookup[address] = peer
                        if len(self.reverse_ip_lookup) > self.reverse_ip_cache_size:
                            self.reverse_ip_lookup.popitem(False)  # Pop the oldest cache entry
                        break
            else:
                # Refresh the peer in the cache (by popping first, it is now on top of the stack again)
                self.reverse_ip_lookup[address] = peer
        return peer

    def get_verified_by_public_key_bin(self, public_key_bin: PublicKeyMat) -> Optional[Peer]:
        """
        Get a verified Peer by its public key bin.
        :param public_key_bin: the string representation of the public key
        :return: the Peer object for this public_key_bin or None
        """
        return self.verified_by_public_key_bin.get(public_key_bin)

    def get_introductions_from(self, peer: Peer) -> List[Address]:
        """
        Get the addresses introduced to us by a certain peer.

        :param peer: the peer to get the introductions for
        :return: a list of the introduced addresses (ip, port)
        """
        introductions = self.reverse_intro_lookup.get(peer, None)
        if introductions is None:
            with self.graph_lock:
                introductions = [k for k, v in self._all_addresses.items()
                                 if v.introduced_by == peer.public_key.key_to_bin()]
                self.reverse_intro_lookup[peer] = introductions
                if len(self.reverse_intro_lookup) > self.reverse_intro_cache_size:
                    self.reverse_intro_lookup.popitem(False)  # Pop the oldest cache entry
        return introductions

    def remove_by_address(self, address: Address) -> None:
        """
        Remove all walkable addresses and verified peers using a certain IP address.

        :param address: the (ip, port) address to remove
        """
        with self.graph_lock:
            self._all_addresses.pop(address, None)
            self.reverse_service_lookup.pop(address, None)
            # Note that the services_per_peer will never be 0, we abuse the lazy `or` to pop the peers from
            # the services_per_peer mapping if they are no longer included. This is fast.
            self.verified_peers = {peer for peer in self.verified_peers
                                   if address not in peer.addresses.values()
                                   or self.services_per_peer.pop(peer.public_key.key_to_bin(), None) == 0}

    def remove_peer(self, peer: Peer) -> None:
        """
        Remove a verified peer.

        :param peer: the Peer to remove
        """
        with self.graph_lock:
            for address in peer.addresses.values():
                self._all_addresses.pop(address, None)
                self.reverse_service_lookup.pop(address, None)
            if peer in self.verified_peers:
                self.verified_peers.remove(peer)
            self.verified_by_public_key_bin.pop(peer.public_key.key_to_bin(), None)
            self.services_per_peer.pop(peer.public_key.key_to_bin(), None)

    def snapshot(self) -> bytes:
        """
        Get a snapshot of all verified peers.

        :return: the serialization (bytes) of all verified peers
        """
        with self.graph_lock:
            out = b""
            for peer in self.verified_peers:
                if peer.address and peer.address != ('0.0.0.0', 0):
                    out += default_serializer.pack('address', peer.address)
            return out

    def load_snapshot(self, snapshot: bytes) -> None:
        """
        Load a snapshot into the walkable addresses.

        This method will prefer returning no peers over throwing an Exception.

        :param snapshot: the snapshot (created by snapshot())
        """
        snaplen = len(snapshot)
        offset = 0
        with self.graph_lock:
            while offset < snaplen:
                previous_offset = offset
                try:
                    address, offset = default_serializer.unpack('address', snapshot, offset)
                    self._all_addresses[address] = WalkableAddress(b'', None, False)
                except Exception:  # pylint: disable=W0703
                    if offset <= previous_offset:
                        # We got stuck, or even went back in time.
                        logging.error("Snapshot loading got stuck! Aborting snapshot load.")
                        break
                    logging.warning("Snapshot failed on entry, skipping %s!", repr(address))



import struct
from binascii import hexlify
from os import urandom

from ..pengbaorange.attestation import create_attest_pair
from ..pengbaorange.structs import PengBaoAttestation
from ..primitives.boneh import generate_keypair
from ..primitives.structs import BonehPrivateKey, BonehPublicKey, pack_pair, unpack_pair
from ...identity_formats import IdentityAlgorithm

LARGE_INTEGER = 32765


def _safe_rndint(key_size, mod):
    """
    Generate a urandom number which is larger than LARGE_INTEGER.

    :param key_size: the bitspace of the number
    :param mod: the modulo
    :return: the random value
    """
    rndint = lambda: int(hexlify(urandom(key_size // 8)), 16) % mod
    out = rndint()
    while out < LARGE_INTEGER:
        out = rndint()
    return out


class PengBaoRangeAlgorithm(IdentityAlgorithm):

    def __init__(self, id_format, formats):
        super(PengBaoRangeAlgorithm, self).__init__(id_format, formats)

        # Check algorithm match
        if formats[id_format]["algorithm"] != "pengbaorange":
            raise RuntimeError("Identity format linked to wrong algorithm")

        # Check key size match
        self.key_size = formats[self.id_format]["key_size"]
        if self.key_size < 32 or self.key_size > 512:
            raise RuntimeError("Illegal key size specified")

        self.a = formats[self.id_format]["min"]
        self.b = formats[self.id_format]["max"]

    def generate_secret_key(self):
        """
        Generate a secret key.

        :return: the secret key
        """
        return generate_keypair(self.key_size)[1]

    def load_secret_key(self, serialized):
        """
        Unserialize a secret key from the key material.

        :param serialized: the string of the private key
        :return: the private key
        """
        return BonehPrivateKey.unserialize(serialized)

    def load_public_key(self, serialized):
        """
        Unserialize a public key from the key material.

        :param serialized: the string of the public key
        :return: the public key
        """
        return BonehPublicKey.unserialize(serialized)

    def get_attestation_class(self):
        """
        Return the Attestation (sub)class for serialization

        :return: the Attestation object
        :rtype: PengBaoAttestation
        """
        return PengBaoAttestation

    def attest(self, PK, value):
        """
        Attest to a value for a certain public key.

        :param PK: the public key of the party we are attesting for
        :type PK: BonehPublicKey
        :param value: the value we are attesting to
        :type value: str
        :return: the attestation string
        :rtype: str
        """
        ivalue = int(hexlify(value), 16)
        return create_attest_pair(PK, ivalue, self.a, self.b, self.key_size).serialize_private(PK)

    def certainty(self, value, aggregate):
        """
        The current certainty of the aggregate object representing a certain value.

        1 is in range, 0 is out of range.

        :param value: the value to match to
        :type value: str
        :param aggregate: the aggregate object
        :type aggregate: dict
        :return: the matching factor [0.0-1.0]
        :rtype: float
        """
        in_range = len(aggregate) > 1
        for k, v in aggregate.items():
            if k != 'attestation':
                in_range &= v
        match = 1.0 if in_range else 0.0
        return match if struct.unpack('>?', value)[0] else 1.0 - match

    def create_challenges(self, PK, attestation):
        """
        Create challenges for a certain counterparty.

        :param PK: the public key of the party we are challenging
        :type PK: BonehPublicKey
        :param attestation: the attestation information
        :type attestation: BonehAttestation
        :return: the challenges to send
        :rtype: [str]
        """
        mod = PK.g.mod - 1
        challenges = [pack_pair(_safe_rndint(self.key_size, mod), _safe_rndint(self.key_size, mod)) for _ in range(1)]
        return challenges

    def create_challenge_response(self, SK, attestation, challenge):
        """
        Create an honest response to a challenge of our value.

        :param SK: our secret key
        :type SK: BonehPrivateKey
        :param attestation: the attestation information
        :type attestation: Attestation
        :param challenge: the challenge to respond to
        :return: the response to a challenge
        :rtype: str
        """
        s, t = unpack_pair(challenge)[0:2]
        # If someone is trying to cheat us by selecting an s or t which is too small, we return random garbage.
        if s < LARGE_INTEGER or t < LARGE_INTEGER:
            return (pack_pair(_safe_rndint(self.key_size, SK.g.mod), _safe_rndint(self.key_size, SK.g.mod))
                    + pack_pair(_safe_rndint(self.key_size, SK.g.mod), _safe_rndint(self.key_size, SK.g.mod)))
        x, y, u, v = attestation.privatedata.generate_response(s, t)
        return pack_pair(x, y) + pack_pair(u, v)

    def create_certainty_aggregate(self, attestation):
        """
        Create an empty aggregate object, for matching to values.

        :return: the aggregate object
        :rtype: dict
        """
        return {'attestation': attestation}

    def create_honesty_challenge(self, PK, value):
        """
        Use a known value to check for honesty.

        :param PK: the public key of the party we are challenging
        :type PK: BonehPublicKey
        :param value: the value to use
        :type value: str
        :return: the challenge to send
        :rtype: str
        """
        raise NotImplementedError()

    def process_honesty_challenge(self, value, response):
        """
        Given a response, check if it matches the expected value.

        :param value: the expected value
        :type value: int
        :param response: the returned response
        :type response: str
        :return: if the value matches the response
        :rtype: bool
        """
        raise NotImplementedError()

    def process_challenge_response(self, aggregate, challenge, response):
        """
        Given a response, update the current aggregate.

        :param aggregate: the aggregate object
        :type aggregate: dict
        :param challenge: the sent challenge
        :type challenge: str
        :param response: the response to introduce
        :type response: str
        :return: the new aggregate
        :rtype: dict
        """
        x, y, rem = unpack_pair(response)
        u, v, _ = unpack_pair(rem)
        s, t, _ = unpack_pair(challenge)
        attestation = aggregate['attestation']
        aggregate[challenge] = attestation.publicdata.check(self.a, self.b, s, t, x, y, u, v)
        return aggregate


__all__ = ["PengBaoAttestation", "PengBaoRangeAlgorithm"]



from unittest.mock import Mock

from ..base import TestBase
from ...dht.provider import DHTCommunityProvider
from ...keyvault.crypto import ECCrypto
from ...messaging.anonymization.tunnel import IntroductionPoint
from ...peer import Peer
from ...util import succeed


class TestNode(TestBase):

    def setUp(self):
        super().setUp()
        self.crypto = ECCrypto()

        self.ip_pk = self.crypto.key_from_public_bin(b'LibNaCLPK:\xc8\xf38};U\xe4\xd5\xf7\xfd\xbc+J!\xbe\xba'
                                                     b'\x81M\xda\xef\xb7\x8c\xacL\x1eZ\x9d\xaf\xaaX+&\xac\xe2'
                                                     b'\xd2\xdd\x86\xa9\x97\xb8T\x9b\x82\xc1>\xa2\r\x11?\xef'
                                                     b'\x137\xf1\xdc!\x7f\x9fW\xe7\x11.\xe2\xc8)')
        self.seeder_pk = self.crypto.key_from_public_bin(b'LibNaCLPK:/N\xc5\xd1#\xd4\xc5\x02\xca\xb4\xa4\xd4vKD'
                                                         b'\xf1"\xf01,\\\xde\x14\x87\xa9\xf6T\x90\xd9\xb0qk\xdbPS'
                                                         b'\xfbqm\xc1,i\xca\x88\x7fm\xe8\\\x0f\xe9\xee\xec\xce\xbeN'
                                                         b'\xdc\x94\xc4\x84\'\x8b\xb8\x8e\x1b\xc4')

        self.intro_point = IntroductionPoint(Peer(self.ip_pk, ('1.2.3.4', 567)),
                                             self.seeder_pk.key_to_bin(), last_seen=0)
        self.info_hash = bytes(range(20))
        self.provider = DHTCommunityProvider(Mock(), self.intro_point.peer.address[1])
        self.dht_value = b'\x01\x01\x02\x03\x04\x027\x00\x00\x00\x00\x00@\xc8\xf38};U\xe4\xd5\xf7\xfd\xbc+J!\xbe' \
                         b'\xba\x81M\xda\xef\xb7\x8c\xacL\x1eZ\x9d\xaf\xaaX+&\xac\xe2\xd2\xdd\x86\xa9\x97\xb8T\x9b' \
                         b'\x82\xc1>\xa2\r\x11?\xef\x137\xf1\xdc!\x7f\x9fW\xe7\x11.\xe2\xc8)\x00@/N\xc5\xd1#\xd4\xc5' \
                         b'\x02\xca\xb4\xa4\xd4vKD\xf1"\xf01,\\\xde\x14\x87\xa9\xf6T\x90\xd9\xb0qk\xdbPS\xfbqm\xc1,i' \
                         b'\xca\x88\x7fm\xe8\\\x0f\xe9\xee\xec\xce\xbeN\xdc\x94\xc4\x84\'\x8b\xb8\x8e\x1b\xc4'

    async def test_announce(self):
        self.provider.dht_community.store_value.return_value = succeed(None)
        await self.provider.announce(self.info_hash, self.intro_point)
        self.provider.dht_community.store_value.assert_called_once_with(self.info_hash, self.dht_value)

    async def test_lookup(self):
        self.provider.dht_community.find_values = lambda _: succeed([(self.dht_value, None)])
        info_hash, intro_points = await self.provider.lookup(self.info_hash)
        assert info_hash == self.info_hash
        assert intro_points[0].peer.address == self.intro_point.peer.address
        assert intro_points[0].peer.public_key.key_to_bin() == self.intro_point.peer.public_key.key_to_bin()
        assert intro_points[0].seeder_pk == self.intro_point.seeder_pk



from binascii import hexlify, unhexlify

from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Boolean, Float, Integer, List, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, HTTP_INTERNAL_SERVER_ERROR, HTTP_NOT_FOUND, Response
from .schema import AddressWithPK, schema
from ..dht.provider import DHTIntroPointPayload
from ..messaging.anonymization.community import (CIRCUIT_STATE_READY, CIRCUIT_TYPE_DATA, IntroductionPoint,
                                                 PEER_FLAG_SPEED_TEST, PEER_SOURCE_DHT, PEER_SOURCE_PEX,
                                                 TunnelCommunity)
from ..messaging.anonymization.utils import run_speed_test
from ..messaging.serialization import PackError
from ..peer import Peer

SpeedTestResponseSchema = schema(SpeedTestResponse={
    "speed": (Float, 'Speed in MiB/s'),
    "messages_sent": Integer,
    "messages_received": Integer,
    "rtt_mean": Float,
    "rtt_median": Float
})


class TunnelEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT data.
    """

    def __init__(self):
        super(TunnelEndpoint, self).__init__()
        self.tunnels = None

    def setup_routes(self):
        self.app.add_routes([web.get('/settings', self.get_settings),
                             web.get('/circuits', self.get_circuits),
                             web.post('/circuits/test', self.speed_test_new_circuit),
                             web.post('/circuits/{circuit_id}/test', self.speed_test_existing_circuit),
                             web.get('/relays', self.get_relays),
                             web.get('/exits', self.get_exits),
                             web.get('/swarms', self.get_swarms),
                             web.get('/swarms/{infohash}/size', self.get_swarm_size),
                             web.get('/peers', self.get_peers),
                             web.get('/peers/dht', self.get_dht_peers),
                             web.get('/peers/pex', self.get_pex_peers)])

    def initialize(self, session):
        super(TunnelEndpoint, self).initialize(session)
        self.tunnels = session.get_overlay(TunnelCommunity)

    @docs(
        tags=["Tunnels"],
        summary="Return a dictionary of all tunnel settings.",
        responses={
            200: {
                "schema": schema(TunnelSettingsResponse={
                    "settings": [schema(TunnelSettings={})]
                })
            }
        }
    )
    def get_settings(self, _):
        return Response({'settings': {k: list(v) if isinstance(v, set) else v
                                      for k, v in self.tunnels.settings.__dict__.items() if k != 'crypto'}})

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all current circuits.",
        responses={
            200: {
                "schema": schema(CircuitsResponse={
                    "circuits": [schema(Circuit={
                        "circuit_id": Integer,
                        "goal_hops": Integer,
                        "actual_hops": Integer,
                        "verified_hops": List(String),
                        "unverified_hop": List(String),
                        "type": String,
                        "state": String,
                        "bytes_up": Integer,
                        "bytes_down": Integer,
                        "creation_time": Integer
                    })]
                })
            }
        }
    )
    async def get_circuits(self, _):
        return Response({"circuits": [{
            "circuit_id": circuit.circuit_id,
            "goal_hops": circuit.goal_hops,
            "actual_hops": len(circuit.hops),
            "verified_hops": [hexlify(hop.mid).decode('utf-8') for hop in circuit.hops],
            "unverified_hop": hexlify(circuit.unverified_hop.mid).decode('utf-8') if circuit.unverified_hop else '',
            "type": circuit.ctype,
            "state": f'{circuit.state} ({circuit.closing_info})' if circuit.closing_info else circuit.state,
            "bytes_up": circuit.bytes_up,
            "bytes_down": circuit.bytes_down,
            "creation_time": circuit.creation_time,
            "exit_flags": circuit.exit_flags
        } for circuit in self.tunnels.circuits.values()]})

    @docs(
        tags=["Tunnels"],
        summary="Test the upload or download speed of a circuit.",
        parameters=[{
            'in': 'path',
            'name': 'circuit_id',
            'description': 'The circuit_id of the circuit which is to be tested.',
            'type': 'integer',
        }],
        responses={
            200: {"schema": SpeedTestResponseSchema}
        }
    )
    @json_schema(schema(SpeedTestRequest={
        'request_size*': (Integer, 'Size of the requests to send (0..1500)'),
        'response_size*': (Integer, 'Size of the responses to send (0..1500)'),
        'num_packets*': (Integer, 'Number of packets to send'),
    }))
    async def speed_test_existing_circuit(self, request):
        if not request.match_info['circuit_id'].isdigit():
            return Response({"error": "circuit_id must be an integer"}, status=HTTP_BAD_REQUEST)

        circuit = self.tunnels.circuits.get(int(request.match_info['circuit_id']))
        if not circuit:
            return Response({"error": "could not find requested circuit"}, status=HTTP_NOT_FOUND)
        if circuit.state != CIRCUIT_STATE_READY:
            return Response({"error": "the requested circuit is not ready to transfer data"}, status=HTTP_BAD_REQUEST)
        if circuit.ctype == CIRCUIT_TYPE_DATA and PEER_FLAG_SPEED_TEST not in circuit.exit_flags:
            return Response({"error": "the requested circuit does not support speed testing"}, status=HTTP_BAD_REQUEST)
        return await self.run_speed_test(circuit, (await request.json()))

    @docs(
        tags=["Tunnels"],
        summary="Test the upload or download speed of a newly created circuit. "
                "The circuit is destroyed after the test has completed.",
        parameters=[{
            'in': 'query',
            'name': 'direction',
            'description': 'The direction for which to test the speed.',
            'type': 'string',
            'enum': ['upload', 'download'],
            'default': 'download'
        }],
        responses={
            200: {"schema": SpeedTestResponseSchema}
        }
    )
    @json_schema(schema(SpeedTestRequest={
        'goals_hops': (Integer, 'Number of hops that the newly created circuit should have'),
        'request_size*': (Integer, 'Size of the requests to send (0..1500)'),
        'response_size*': (Integer, 'Size of the responses to send (0..1500)'),
        'num_packets*': (Integer, 'Number of packets to send'),
    }))
    async def speed_test_new_circuit(self, request):
        params = await request.json()

        goal_hops = params.get('goal_hops', 1)
        if not 1 <= goal_hops <= 3:
            return Response({"error": "invalid number of hops specified"}, status=HTTP_BAD_REQUEST)

        circuit = self.tunnels.create_circuit(goal_hops, ctype='SPEED_TEST', exit_flags=(PEER_FLAG_SPEED_TEST,))
        if not circuit or not await circuit.ready:
            return Response({"error": "failed to create circuit"}, status=HTTP_INTERNAL_SERVER_ERROR)

        result = await self.run_speed_test(circuit, params)
        self.tunnels.remove_circuit(circuit.circuit_id, additional_info='speed test finished')
        return result

    async def run_speed_test(self, circuit, params):
        if not 0 <= params.get('request_size', -1) <= 2000 or not 0 <= params.get('response_size', -1) <= 2000:
            return Response({"error": "invalid request or response size specified"}, status=HTTP_BAD_REQUEST)
        if 'num_requests' not in params:
            return Response({"error": "number of requests is not specified"}, status=HTTP_BAD_REQUEST)
        return Response(await run_speed_test(self.tunnels, circuit,
                                             params['request_size'], params['response_size'], params['num_requests']))

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all current relays.",
        responses={
            200: {
                "schema": schema(RelaysResponse={
                    "relays": [schema(Relay={
                        "circuit_from": Integer,
                        "circuit_to": Integer,
                        "is_rendezvous": Boolean,
                        "bytes_up": Integer,
                        "bytes_down": Integer,
                        "creation_time": Integer
                    })]
                })
            }
        }
    )
    async def get_relays(self, _):
        return Response({"relays": [{
            "circuit_from": circuit_from,
            "circuit_to": relay.circuit_id,
            "is_rendezvous": relay.rendezvous_relay,
            "bytes_up": relay.bytes_up,
            "bytes_down": relay.bytes_down,
            "creation_time": relay.creation_time
        } for circuit_from, relay in self.tunnels.relay_from_to.items()]})

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all current exits.",
        responses={
            200: {
                "schema": schema(ExitsResponse={
                    "exits": [schema(Exit={
                        "circuit_from": Integer,
                        "enabled": Boolean,
                        "bytes_up": Integer,
                        "bytes_down": Integer,
                        "creation_time": Integer
                    })]
                })
            }
        }
    )
    async def get_exits(self, _):
        return Response({"exits": [{
            "circuit_from": circuit_from,
            "enabled": exit_sock.enabled,
            "bytes_up": exit_sock.bytes_up,
            "bytes_down": exit_sock.bytes_down,
            "creation_time": exit_sock.creation_time,
            "is_introduction": exit_sock.circuit_id in [c.circuit_id for c, _ in self.tunnels.intro_point_for.values()],
            "is_rendezvous": exit_sock.circuit_id in [c.circuit_id for c in self.tunnels.rendezvous_point_for.values()]
        } for circuit_from, exit_sock in self.tunnels.exit_sockets.items()]})

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all current hidden swarms.",
        responses={
            200: {
                "schema": schema(SwarmsResponse={
                    "swarms": [schema(Swarm={
                        "info_hash": String,
                        "num_seeders": Integer,
                        "num_connections": Integer,
                        "num_connections_incomplete": Integer,
                        "seeding": Boolean,
                        "last_lookup": Integer,
                        "bytes_up": Integer,
                        "bytes_down": Integer
                    })]
                })
            }
        }
    )
    async def get_swarms(self, _):
        return Response({"swarms": [{
            "info_hash": hexlify(swarm.info_hash).decode('utf-8'),
            "num_seeders": swarm.get_num_seeders(),
            "num_connections": swarm.get_num_connections(),
            "num_connections_incomplete": swarm.get_num_connections_incomplete(),
            "num_ips_from_dht": len([ip for ip in swarm.intro_points if ip.source == PEER_SOURCE_DHT]),
            "num_ips_from_pex": len([ip for ip in swarm.intro_points if ip.source == PEER_SOURCE_PEX]),
            "seeding": swarm.seeding,
            "last_lookup": swarm.last_lookup,
            "bytes_up": swarm.get_total_up(),
            "bytes_down": swarm.get_total_down()
        } for swarm in self.tunnels.swarms.values()]})

    @docs(
        tags=["Tunnels"],
        summary="Estimate the hidden swarm size for a given infohash.",
        parameters=[{
            'in': 'path',
            'name': 'infohash',
            'description': 'Infohash of the swarm for which to estimate the size.',
            'type': 'string',
            'required': True
        }],
        responses={
            200: {
                "schema": schema(SwarmsSizeResponse={
                    "swarm_size": Integer
                })
            }
        }
    )
    async def get_swarm_size(self, request):
        infohash = unhexlify(request.match_info['infohash'])
        swarm_size = await self.tunnels.estimate_swarm_size(infohash, hops=request.query.get('hops', 1))
        return Response({"swarm_size": swarm_size})

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all peers currently part of the tunnel community.",
        responses={
            200: {
                "schema": schema(TunnelPeersResponse={
                    "peers": [schema(TunnelPeer={
                        "ip": String,
                        "port": Integer,
                        "mid": String,
                        "is_key_compatible": Boolean,
                        "flags": List(Integer),
                    })]
                })
            }
        }
    )
    async def get_peers(self, _):
        return Response({"peers": [{
            "ip": peer.address[0],
            "port": peer.address[1],
            "mid": hexlify(peer.mid).decode('utf-8'),
            "is_key_compatible": self.tunnels.crypto.is_key_compatible(peer.public_key),
            "flags": flags
        } for peer, flags in self.tunnels.candidates.items()]})

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all hidden services peers that are in the local DHT store.",
        responses={
            200: {
                "schema": schema(TunnelDHTPeersResponse={
                    "peers": [AddressWithPK]
                })
            }
        }
    )
    async def get_dht_peers(self, _):
        dht = self.tunnels.dht_provider.dht_community
        ips_by_infohash = {}
        for storage in dht.storages.values():
            for key, raw_values in storage.items.items():
                ips_by_infohash[key] = []
                for value in dht.post_process_values([v.data for v in raw_values]):
                    try:
                        payload, _ = dht.serializer.unpack_serializable(DHTIntroPointPayload, value[0])
                    except PackError:
                        continue
                    peer = Peer(b'LibNaCLPK:' + payload.intro_pk, payload.address)
                    ip = IntroductionPoint(peer, b'LibNaCLPK:' + payload.seeder_pk, PEER_SOURCE_DHT, payload.last_seen)
                    ips_by_infohash[key].append(ip)

        return Response([{'info_hash': hexlify(h).decode(),
                          'peers': [i.to_dict() for i in ips]} for h, ips in ips_by_infohash.items()])

    @docs(
        tags=["Tunnels"],
        summary="Return a list of all hidden services peers that are in the local PEX store.",
        responses={
            200: {
                "schema": schema(TunnelPEXPeersResponse={
                    "peers": [AddressWithPK]
                })
            }
        }
    )
    async def get_pex_peers(self, _):
        return Response([{'info_hash': hexlify(h).decode(),
                          'peers': [i.to_dict() for i in c.get_intro_points()]} for h, c in self.tunnels.pex.items()])



import time

from .base import TestDHTBase
from ...dht.churn import PingChurn
from ...dht.community import DHTCommunity


class TestPingChurn(TestDHTBase):

    def setUp(self):
        super(TestPingChurn, self).setUp()
        self.initialize(DHTCommunity, 2)

        self.strategies = []
        for i in range(2):
            self.strategies.append(PingChurn(self.overlay(i), ping_interval=0.0))

    async def test_ping_all(self):
        await self.introduce_nodes()
        bucket = self.routing_table(0).trie[u'']

        node1 = bucket.get(self.my_node_id(1))
        node1.failed = 1
        node1.last_response = 0

        self.strategies[0].take_step()
        await self.deliver_messages()

        self.assertTrue(node1.failed == 0)
        self.assertNotEqual(node1.last_response, 0)

    async def test_ping_all_skip(self):
        await self.introduce_nodes()
        bucket = self.routing_table(0).trie[u'']
        node1 = bucket.get(self.my_node_id(1))
        node1.failed = 1
        node1.last_response = time.time() + 5

        self.strategies[0].take_step()
        self.assertTrue(node1.failed == 1)



from .base import TestBase
from ..configuration import (Bootstrapper, BootstrapperDefinition, ConfigBuilder, DISPERSY_BOOTSTRAPPER, Strategy,
                             WalkerDefinition, get_default_configuration)


class TestConfiguration(TestBase):

    def assertDictInDict(self, expected: dict, container: dict):
        self.assertTrue(any(all(entry.get(key) == expected[key] for key in expected.keys())
                            and len(entry) == len(expected)
                            for entry in container))

    def test_clear_keys(self):
        """
        Check if all keys are cleared when requested.
        """
        builder = ConfigBuilder().clear_keys()

        self.assertEqual(0, len(builder.config['keys']))

    def test_clear_overlays(self):
        """
        Check if all overlays are cleared when requested.
        """
        builder = ConfigBuilder().clear_overlays()

        self.assertEqual(0, len(builder.finalize()['overlays']))

    def test_add_illegal_port_negative(self):
        """
        Check if negative ports raise an error on finalization.
        """
        builder = ConfigBuilder().set_port(-1)

        self.assertRaises(AssertionError, builder.finalize)

    def test_add_illegal_port_big(self):
        """
        Check if ports over 65535 raise an error on finalization.
        """
        builder = ConfigBuilder().set_port(100000)

        self.assertRaises(AssertionError, builder.finalize)

    def test_change_port(self):
        """
        Check if changes to the port are finalized.
        """
        builder = ConfigBuilder().set_port(1000)

        self.assertEqual(1000, builder.finalize()['port'])

    def test_add_illegal_address(self):
        """
        Check if wrong IP addresses raise an error immediately.
        """
        builder = ConfigBuilder()

        self.assertRaises(AssertionError, builder.set_address, "1.1.1")

    def test_change_address(self):
        """
        Check if changes to the address are finalized.
        """
        builder = ConfigBuilder().set_address("1.1.1.1")

        self.assertEqual("1.1.1.1", builder.finalize()['address'])

    def test_set_illegal_log_level(self):
        """
        Check if wrong log levels raise an error immediately.
        """
        builder = ConfigBuilder()

        self.assertRaises(AssertionError, builder.set_log_level, "I don't exist")

    def test_change_log_level(self):
        """
        Check if changes to the log level are finalized.
        """
        builder = ConfigBuilder().set_log_level("CRITICAL")

        self.assertEqual("CRITICAL", builder.finalize()['logger']['level'])

    def test_set_illegal_walk_interval(self):
        """
        Check if negative walk intervals raise an error on finalization.
        """
        builder = ConfigBuilder().set_walker_interval(-1.0)

        self.assertRaises(AssertionError, builder.finalize)

    def test_change_walk_interval(self):
        """
        Check if changes to the walk interval are finalized.
        """
        builder = ConfigBuilder().set_walker_interval(3.14)

        self.assertEqual(3.14, builder.finalize()['walker_interval'])

    def test_add_key_illegal_curve(self):
        """
        Check if wrong key curves raise an error immediately.
        """
        builder = ConfigBuilder()

        self.assertRaises(AssertionError, builder.add_key, "my key", "I don't exist", "some file")

    def test_add_key(self):
        """
        Check if changes to the keys are finalized.
        """
        builder = ConfigBuilder().add_key("my new key", "very-low", "some file")

        expected = {
            'alias': "my new key",
            'generation': "very-low",
            'file': "some file"
        }
        keys = builder.finalize()['keys']

        self.assertEqual(1 + len(get_default_configuration()['keys']), len(keys))
        self.assertTrue(any(set(entry.items()) == set(expected.items()) for entry in keys))

    def test_add_overlay_overwrite(self):
        """
        Check if the allow duplicate flag does not introduce duplicates.
        """
        builder = ConfigBuilder().add_overlay("DiscoveryCommunity", "anonymous id", [], [], {}, [],
                                              allow_duplicate=False)

        expected = {
            'class': "DiscoveryCommunity",
            'key': "anonymous id",
            'walkers': [],
            'bootstrappers': [],
            'initialize': {},
            'on_start': []
        }

        self.assertEqual(len(get_default_configuration()['overlays']), len(builder.finalize()['overlays']))
        self.assertDictInDict(expected, builder.finalize()['overlays'])

    def test_add_overlay_append(self):
        """
        Check if the duplicate overlays are simply appended.
        """
        builder = ConfigBuilder().add_overlay("DiscoveryCommunity", "anonymous id", [], [], {}, [],
                                              allow_duplicate=True)

        expected = {
            'class': "DiscoveryCommunity",
            'key': "anonymous id",
            'walkers': [],
            'bootstrappers': [],
            'initialize': {},
            'on_start': []
        }

        self.assertEqual(1 + len(get_default_configuration()['overlays']), len(builder.finalize()['overlays']))
        self.assertDictInDict(expected, builder.finalize()['overlays'])

    def test_add_overlay_complex(self):
        """
        Check if a complex overlay is correctly added.
        """
        builder = ConfigBuilder().add_overlay("MyCommunity",
                                              "anonymous id",
                                              [WalkerDefinition(Strategy.RandomWalk, 42, {'timeout': 3.0})],
                                              [BootstrapperDefinition(Bootstrapper.DispersyBootstrapper,
                                                                      {'ip_addresses': [('1.2.3.4', 5)],
                                                                       'dns_addresses': [('tribler.org', 5)]})],
                                              {'settings': {"my_key": "my_value"}},
                                              [('do_a_thing', 42), ('do_another_thing', )])

        expected = {
            'class': "MyCommunity",
            'key': "anonymous id",
            'walkers': [{
                'strategy': "RandomWalk",
                'peers': 42,
                'init': {
                    'timeout': 3.0
                }
            }],
            'bootstrappers': [{
                'class': "DispersyBootstrapper",
                'init': {
                    'ip_addresses': [('1.2.3.4', 5)],
                    'dns_addresses': [('tribler.org', 5)]
                }
            }],
            'initialize': {'settings': {"my_key": "my_value"}},
            'on_start': [('do_a_thing', 42), ('do_another_thing', )]
        }

        self.assertDictInDict(expected, builder.finalize()['overlays'])

    def test_illegal_random_churn_strategy(self):
        """
        Only the DiscoveryCommunity may use the RandomChurn strategy.
        """
        builder = ConfigBuilder().add_overlay("MyCommunity",
                                              "anonymous id",
                                              [WalkerDefinition(Strategy.RandomChurn, 20, {})],
                                              [],
                                              {},
                                              [])

        self.assertRaises(AssertionError, builder.finalize)

    def test_illegal_periodic_similarity_strategy(self):
        """
        Only the DiscoveryCommunity may use the PeriodicSimilarity strategy.
        """
        builder = ConfigBuilder().add_overlay("MyCommunity",
                                              "anonymous id",
                                              [WalkerDefinition(Strategy.PeriodicSimilarity, 20, {})],
                                              [],
                                              {},
                                              [])

        self.assertRaises(AssertionError, builder.finalize)

    def test_correct_random_churn_strategy(self):
        """
        The DiscoveryCommunity may use the RandomChurn strategy.
        """
        builder = ConfigBuilder().add_overlay("DiscoveryCommunity",
                                              "anonymous id",
                                              [WalkerDefinition(Strategy.RandomChurn, 20, {})],
                                              [],
                                              {},
                                              [])

        expected = {
            'class': "DiscoveryCommunity",
            'key': "anonymous id",
            'walkers': [{
                'strategy': "RandomChurn",
                'peers': 20,
                'init': {}
            }],
            'bootstrappers': [],
            'initialize': {},
            'on_start': []
        }

        self.assertEqual(len(get_default_configuration()['overlays']), len(builder.finalize()['overlays']))
        self.assertDictInDict(expected, builder.finalize()['overlays'])

    def test_correct_periodic_similarity_strategy(self):
        """
        The DiscoveryCommunity may use the PeriodicSimilarity strategy.
        """
        builder = ConfigBuilder().add_overlay("DiscoveryCommunity",
                                              "anonymous id",
                                              [WalkerDefinition(Strategy.PeriodicSimilarity, 20, {})],
                                              [],
                                              {},
                                              [])

        expected = {
            'class': "DiscoveryCommunity",
            'key': "anonymous id",
            'walkers': [{
                'strategy': "PeriodicSimilarity",
                'peers': 20,
                'init': {}
            }],
            'bootstrappers': [],
            'initialize': {},
            'on_start': []
        }

        self.assertEqual(len(get_default_configuration()['overlays']), len(builder.finalize()['overlays']))
        self.assertDictInDict(expected, builder.finalize()['overlays'])

    def test_default_configuration(self):
        """
        Check if we can reconstruct the default configuration.
        """
        builder = ConfigBuilder(True).set_address("0.0.0.0") \
                                     .set_port(8090) \
                                     .add_key("anonymous id", "curve25519", "ec_multichain.pem") \
                                     .set_log_level("INFO") \
                                     .set_working_directory(".") \
                                     .set_walker_interval(0.5) \
                                     .add_overlay("DiscoveryCommunity",
                                                  "anonymous id",
                                                  [WalkerDefinition(Strategy.RandomWalk, 20, {'timeout': 3.0}),
                                                   WalkerDefinition(Strategy.RandomChurn, -1, {
                                                       'sample_size': 8,
                                                       'ping_interval': 10.0,
                                                       'inactive_time': 27.5,
                                                       'drop_time': 57.5
                                                   }),
                                                   WalkerDefinition(Strategy.PeriodicSimilarity, -1, {})],
                                                  [BootstrapperDefinition(Bootstrapper.DispersyBootstrapper,
                                                                          DISPERSY_BOOTSTRAPPER['init'])],
                                                  {},
                                                  []) \
                                     .add_overlay("HiddenTunnelCommunity",
                                                  "anonymous id",
                                                  [WalkerDefinition(Strategy.RandomWalk, 20, {'timeout': 3.0})],
                                                  [BootstrapperDefinition(Bootstrapper.DispersyBootstrapper,
                                                                          DISPERSY_BOOTSTRAPPER['init'])],
                                                  {'settings': {
                                                      'min_circuits': 1,
                                                      'max_circuits': 1,
                                                      'max_relays_or_exits': 100,
                                                      'max_time': 10 * 60,
                                                      'max_time_inactive': 20,
                                                      'max_traffic': 250 * 1024 * 1024,
                                                      'dht_lookup_interval': 30
                                                  }},
                                                  [('build_tunnels', 1)]) \
                                     .add_overlay("DHTDiscoveryCommunity",
                                                  "anonymous id",
                                                  [WalkerDefinition(Strategy.RandomWalk, 20, {'timeout': 3.0}),
                                                   WalkerDefinition(Strategy.PingChurn, -1, {})],
                                                  [BootstrapperDefinition(Bootstrapper.DispersyBootstrapper,
                                                                          DISPERSY_BOOTSTRAPPER['init'])],
                                                  {},
                                                  [])

        self.assertDictEqual(get_default_configuration(), builder.finalize())



import struct

from ..base import TestBase
from ...messaging.serialization import DefaultStruct, PackError, Serializable, Serializer


class Short(Serializable):
    format_list = ["H"]

    def __init__(self, number):
        self.number = number

    def to_pack_list(self):
        return [("H", self.number)]

    @classmethod
    def from_unpack_list(cls, *args):
        return Short(*args)


class Byte(Serializable):
    format_list = ["B"]

    def __init__(self, byte):
        self.byte = byte

    def to_pack_list(self):
        return [("B", self.byte)]

    @classmethod
    def from_unpack_list(cls, *args):
        return Byte(*args)


class Nested(Serializable):
    format_list = [[Byte]]

    def __init__(self, byte_list):
        self.byte_list = byte_list

    def to_pack_list(self):
        return [('payload-list', self.byte_list)]

    @classmethod
    def from_unpack_list(cls, *args):
        return Nested(*args)


class TestSerializer(TestBase):

    def setUp(self):
        super(TestSerializer, self).setUp()
        self.serializer = Serializer()

    def check_pack_unpack(self, format_ser, format_unser, value):
        packer = self.serializer.get_packer_for(format_ser)
        values = (value,) if not isinstance(value, (list, tuple)) else value
        serialized = packer.pack(*values)

        unpack_list = []
        packer = self.serializer.get_packer_for(format_unser)
        packer.unpack(serialized, 0, unpack_list)

        self.assertEqual(value, unpack_list[0])

    def test_pack_bool_true(self):
        """
        Check if 'true' booleans can be correctly packed and unpacked.
        """
        self.check_pack_unpack('?', '?', True)

    def test_pack_bool_false(self):
        """
        Check if 'false' booleans can be correctly packed and unpacked.
        """
        self.check_pack_unpack('?', '?', False)

    def test_pack_byte_0(self):
        """
        Check if a 0 (unsigned byte) can be correctly packed and unpacked.
        """
        self.check_pack_unpack('B', 'B', 0)

    def test_pack_byte_1(self):
        """
        Check if a 1 (unsigned byte) can be correctly packed and unpacked.
        """
        self.check_pack_unpack('B', 'B', 1)

    def test_pack_byte_255(self):
        """
        Check if a 255 (unsigned byte) can be correctly packed and unpacked.
        """
        self.check_pack_unpack('B', 'B', 255)

    def test_pack_byte_256(self):
        """
        Check if a 256 (unsigned byte) throws a struct.error.
        """
        self.assertRaises(struct.error, self.check_pack_unpack, 'B', 'B', 256)

    def test_unpack_short_truncated(self):
        """
        Check if 1 byte string cannot be unpacked as a short.
        """
        self.assertRaises(struct.error, self.check_pack_unpack, 'B', 'H', 255)

    def test_pack_list(self):
        """
        Check if a list of shorts is correctly packed and unpacked.
        """
        self.check_pack_unpack('HH', 'HH', (0, 1337))

    def test_get_formats(self):
        """
        Check if all reported formats contain valid packers.
        """
        formats = self.serializer.get_available_formats()

        for format in formats:
            packer = self.serializer.get_packer_for(format)
            pack_name = "%s(%s)" % (packer.__class__.__name__, format)
            self.assertTrue(hasattr(packer, 'pack'), msg='%s has no pack() method' % pack_name)
            self.assertTrue(callable(getattr(packer, 'pack')), msg='%s.pack is not a method' % pack_name)
            self.assertTrue(hasattr(packer, 'unpack'), msg='%s has no unpack() method' % pack_name)
            self.assertTrue(callable(getattr(packer, 'unpack')), msg='%s.unpack is not a method' % pack_name)

    def test_add_packer(self):
        """
        Check if we can add a packer on the fly.
        """
        self.serializer.add_packer("H_LE", DefaultStruct("<H"))  # little-endian

        serialized = self.serializer.get_packer_for("H_LE").pack(1)  # Packed as 01 00

        unpacked = []
        self.serializer.get_packer_for("H_LE").unpack(serialized, 0, unpacked)  # little-endian, unpacked as 00 01 = 1
        self.serializer.get_packer_for("H").unpack(serialized, 0, unpacked)  # big-endian, unpacked as 01 00 = 256

        self.assertEqual([1, 256], unpacked)

    def test_nested_serializable(self):
        """
        Check if we can unpack nested serializables.
        """
        instance = Short(123)

        data = self.serializer.pack_serializable(instance)
        output, _ = self.serializer.unpack_serializable(Short, data)

        self.assertEqual(instance.number, output.number)

    def test_serializable_byte_256(self):
        """
        Check if pack_serializable of a 256 (unsigned byte) raises a PackError.
        """
        self.assertRaises(PackError, self.serializer.pack_serializable, Byte(256))

    def test_serializable_short_from_byte(self):
        """
        Check if a unpack_serializable of a short from a byte raises a PackError.
        """
        serialized = self.serializer.pack_serializable(Byte(1))
        self.assertRaises(PackError, self.serializer.unpack_serializable, Short, serialized)

    def test_serializable_list(self):
        """
        Check if we can (un)pack serializables easily.
        """
        instance1 = Short(123)
        instance2 = Short(456)

        data = self.serializer.pack_serializable_list([instance1, instance2])
        deserialized = self.serializer.unpack_serializable_list([Short, Short], data)

        self.assertEqual(instance1.number, 123)
        self.assertEqual(instance1.number, deserialized[0].number)
        self.assertEqual(instance2.number, 456)
        self.assertEqual(instance2.number, deserialized[1].number)

    def test_serializable_list_extra_data(self):
        """
        Check if we throw an error when we have too much data to unpack.
        """
        instance1 = Short(123)
        instance2 = Short(456)

        data = self.serializer.pack_serializable_list([instance1, instance2])
        self.assertRaises(PackError, self.serializer.unpack_serializable_list, [Short, Short], data + b"Nope.avi")

    def test_nested_payload_list(self):
        serializable = Nested([Byte(1), Byte(2)])
        data = self.serializer.pack_serializable(serializable)
        decoded, _ = self.serializer.unpack_serializable(Nested, data)
        self.assertEqual(serializable.byte_list[0].byte, decoded.byte_list[0].byte)
        self.assertEqual(serializable.byte_list[1].byte, decoded.byte_list[1].byte)



import json
import logging
import typing

from aiohttp import web

HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_NOT_FOUND = 404
HTTP_CONFLICT = 409
HTTP_PRECONDITION_FAILED = 412
HTTP_INTERNAL_SERVER_ERROR = 500

DEFAULT_HEADERS: typing.Dict[str, str] = {}


class BaseEndpoint:

    def __init__(self, middlewares=()):
        self._logger = logging.getLogger(self.__class__.__name__)
        self.app = web.Application(middlewares=middlewares)
        self.session = None
        self.endpoints = {}
        self.setup_routes()

    def setup_routes(self):
        pass

    def initialize(self, session):
        self.session = session
        for endpoint in self.endpoints.values():
            endpoint.initialize(session)

    def add_endpoint(self, prefix, endpoint):
        self.endpoints[prefix] = endpoint
        self.app.add_subapp(prefix, endpoint.app)


class Response(web.Response):

    def __init__(self, body=None, headers=None, content_type=None, status=200, **kwargs):
        if isinstance(body, (dict, list)):
            body = json.dumps(body)
            content_type = 'application/json'
        super(Response, self).__init__(body=body, headers=headers or DEFAULT_HEADERS,
                                       content_type=content_type, status=status, **kwargs)



from __future__ import annotations

import copy
import enum
import socket
import typing
from typing import Any, Dict

DISPERSY_BOOTSTRAPPER: Dict[Any, Any] = {
    'class': "DispersyBootstrapper",
    'init': {
        'ip_addresses': [
            ("130.161.119.206", 6421),
            ("130.161.119.206", 6422),
            ("131.180.27.155", 6423),
            ("131.180.27.156", 6424),
            ("131.180.27.161", 6427),
            ("131.180.27.161", 6521),
            ("131.180.27.161", 6522),
            ("131.180.27.162", 6523),
            ("131.180.27.162", 6524),
            ("130.161.119.215", 6525),
            ("130.161.119.215", 6526),
            ("130.161.119.201", 6527),
            ("130.161.119.201", 6528)
        ],
        'dns_addresses': [
            (u"dispersy1.tribler.org", 6421), (u"dispersy1.st.tudelft.nl", 6421),
            (u"dispersy2.tribler.org", 6422), (u"dispersy2.st.tudelft.nl", 6422),
            (u"dispersy3.tribler.org", 6423), (u"dispersy3.st.tudelft.nl", 6423),
            (u"dispersy4.tribler.org", 6424),
            (u"tracker1.ip-v8.org", 6521),
            (u"tracker2.ip-v8.org", 6522),
            (u"tracker3.ip-v8.org", 6523),
            (u"tracker4.ip-v8.org", 6524),
            (u"tracker5.ip-v8.org", 6525),
            (u"tracker6.ip-v8.org", 6526),
            (u"tracker7.ip-v8.org", 6527),
            (u"tracker8.ip-v8.org", 6528)
        ],
        'bootstrap_timeout': 30.0
    }
}

default = {
    'address': '0.0.0.0',
    'port': 8090,
    'keys': [
        {
            'alias': "anonymous id",
            'generation': u"curve25519",
            'file': u"ec_multichain.pem"
        }
    ],
    'logger': {
        'level': "INFO"
    },
    'working_directory': ".",
    'walker_interval': 0.5,
    'overlays': [
        {
            'class': 'DiscoveryCommunity',
            'key': "anonymous id",
            'walkers': [
                {
                    'strategy': "RandomWalk",
                    'peers': 20,
                    'init': {
                        'timeout': 3.0
                    }
                },
                {
                    'strategy': "RandomChurn",
                    'peers': -1,
                    'init': {
                        'sample_size': 8,
                        'ping_interval': 10.0,
                        'inactive_time': 27.5,
                        'drop_time': 57.5
                    }
                },
                {
                    'strategy': "PeriodicSimilarity",
                    'peers': -1,
                    'init': {}
                }
            ],
            'bootstrappers': [DISPERSY_BOOTSTRAPPER.copy()],
            'initialize': {},
            'on_start': []
        },
        {
            'class': 'HiddenTunnelCommunity',
            'key': "anonymous id",
            'walkers': [
                {
                    'strategy': "RandomWalk",
                    'peers': 20,
                    'init': {
                        'timeout': 3.0
                    }
                }
            ],
            'bootstrappers': [DISPERSY_BOOTSTRAPPER.copy()],
            'initialize': {
                'settings': {
                    'min_circuits': 1,
                    'max_circuits': 1,
                    'max_relays_or_exits': 100,
                    'max_time': 10 * 60,
                    'max_time_inactive': 20,
                    'max_traffic': 250 * 1024 * 1024,
                    'dht_lookup_interval': 30
                }
            },
            'on_start': [
                ('build_tunnels', 1)
            ]
        },
        {
            'class': 'DHTDiscoveryCommunity',
            'key': "anonymous id",
            'walkers': [
                {
                    'strategy': "RandomWalk",
                    'peers': 20,
                    'init': {
                        'timeout': 3.0
                    }
                },
                {
                    'strategy': "PingChurn",
                    'peers': -1,
                    'init': {}
                }
            ],
            'bootstrappers': [DISPERSY_BOOTSTRAPPER.copy()],
            'initialize': {},
            'on_start': []
        }
    ]
}


def get_default_configuration():
    return copy.deepcopy(default)


class Strategy(enum.Enum):
    RandomWalk = "RandomWalk"
    RandomChurn = "RandomChurn"
    PeriodicSimilarity = "PeriodicSimilarity"
    EdgeWalk = "EdgeWalk"
    PingChurn = "PingChurn"

    @classmethod
    def values(cls):
        return [name for name, _ in cls.__members__.items()]


class WalkerDefinition(typing.NamedTuple):
    strategy: Strategy
    peers: int
    init: dict


class Bootstrapper(enum.Enum):
    DispersyBootstrapper = "DispersyBootstrapper"
    UDPBroadcastBootstrapper = "UDPBroadcastBootstrapper"

    @classmethod
    def values(cls):
        return [name for name, _ in cls.__members__.items()]


class BootstrapperDefinition(typing.NamedTuple):
    bootstrapper: Bootstrapper
    init: dict


default_bootstrap_defs = [BootstrapperDefinition(Bootstrapper.DispersyBootstrapper, DISPERSY_BOOTSTRAPPER['init'])]


class ConfigBuilder(object):

    def __init__(self, clean: bool = False):
        self.config = {} if clean else get_default_configuration()

    def finalize(self) -> dict:
        """
        Process the builder input and check for errors, this produces a configuration dictionary suitable for IPv8.

        Essentially this only checks the `config` variable for errors, use that instead if you're confident you never
        make any errors.
        """
        assert self.config.get('address') is not None, "Missing address in config!"
        assert self.config.get('port') is not None, "Missing port in config!"
        assert self.config.get('keys') is not None, "Missing keys in config!"
        assert self.config.get('logger') is not None, "Missing logger in config!"
        assert self.config.get('walker_interval') is not None, "Missing walker_interval in config!"
        assert self.config.get('overlays') is not None, "Missing overlays in config!"
        assert self.config.get('working_directory') is not None, "Missing working_directory in config!"

        socket.inet_aton(self.config.get('address'))  # Errors out if the address is illegal
        assert 0 <= self.config['port'] <= 65535
        assert self.config['walker_interval'] >= 0

        for overlay in self.config['overlays']:
            assert overlay.get('class') is not None, "Missing class in overlay config!"
            assert overlay.get('key') is not None, f"Missing key in overlay config of {overlay['class']}!"
            assert overlay.get('walkers') is not None, f"Missing walkers in overlay config of {overlay['class']}!"
            assert overlay.get('bootstrappers') is not None,\
                f"Missing bootstrappers in overlay config of {overlay['class']}!"
            assert overlay.get('initialize') is not None,\
                f"Missing initialize in overlay config of {overlay['class']}!"
            assert overlay.get('on_start') is not None, f"Missing on_start in overlay config of {overlay['class']}!"
            assert any(key['alias'] == overlay['key'] for key in self.config['keys']),\
                f"Unknown key alias {overlay['key']} in overlay config of {overlay['class']}!"
            assert all(isinstance(key, str) for key in overlay['initialize'].keys()),\
                f"Keyword argument mapping keys must be strings in overlay config of {overlay['class']}!"
            assert all(isinstance(entry[0], str) for entry in overlay['on_start']),\
                f"Start methods must be strings in overlay config of {overlay['class']}!"
            for walker in overlay['walkers']:
                assert walker.get('strategy') is not None,\
                    f"Missing strategy class in strategy config of {overlay['class']}!"
                assert walker.get('peers') is not None,\
                    f"Missing peers in {walker['strategy']} config of {overlay['class']}!"
                assert walker.get('init') is not None,\
                    f"Missing init in {walker['strategy']} config of {overlay['class']}!"
                assert walker['strategy'] in Strategy.values()
                if (walker['strategy'] == Strategy.RandomChurn.value
                        or walker['strategy'] == Strategy.PeriodicSimilarity.value):
                    assert overlay['class'] == 'DiscoveryCommunity'
            for bootstrapper in overlay['bootstrappers']:
                assert bootstrapper.get('class') is not None,\
                    f"Missing bootstrapper class in bootstrapper config of {overlay['class']}!"
                assert bootstrapper.get('init') is not None,\
                    f"Missing init in {bootstrapper['class']} config of {overlay['class']}!"
                assert bootstrapper['class'] in Bootstrapper.values()

        return self.config

    def clear_keys(self) -> ConfigBuilder:
        """
        Remove all keys in the current configuration.
        """
        self.config['keys'] = []
        return self

    def clear_overlays(self) -> ConfigBuilder:
        """
        Remove all overlays in the current configuration.
        """
        self.config['overlays'] = []
        return self

    def set_address(self, address: str) -> ConfigBuilder:
        """
        Set the address IPv8 is to try and bind to.

        For localhost only use: 127.0.0.1
        For Internet communication use: 0.0.0.0
        """
        assert address.count('.') == 3
        self.config['address'] = address
        return self

    def set_port(self, port: int) -> ConfigBuilder:
        """
        Set the port that IPv8 should TRY to bind to.
        If your port is not available, IPv8 will try and find another one.
        """
        self.config['port'] = port
        return self

    def set_log_level(self, log_level: str) -> ConfigBuilder:
        """
        Set the log level for all of IPv8's loggers.
        Choose from 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG' or 'NOTSET'.
        """
        assert log_level in ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET']
        self.config['logger'] = {'level': log_level}
        return self

    def set_working_directory(self, folder_path: str) -> ConfigBuilder:
        """
        Set the common working directory for overlays.

        Individual settings may override this working directory, but it serves as the base working directory.
        """
        self.config['working_directory'] = folder_path
        return self

    def set_walker_interval(self, interval: float) -> ConfigBuilder:
        """
        Set the interval, in seconds, at which IPv8 schedules its strategies.

        An interval of 3.14 would mean that each walker attempts to walk every 3.14 seconds.
        Setting this interval too low will choke the event thread and decrease the QoS of IPv8 overlays.
        Setting this interval too high will decrease the QoS of IPv8 overlays due to lack of content.
        """
        self.config['walker_interval'] = interval
        return self

    def add_key(self, alias: str, generation: str, file_path: str) -> ConfigBuilder:
        """
        Add a key by alias and mode of generation, to be stored at a certain file path.

        If a key already exists at the given file path, that will be loaded instead of generating a new key.
        """
        assert generation in ['curve25519', 'very-low', 'low', 'medium', 'high']
        if 'keys' in self.config:
            self.config['keys'] = [key for key in self.config['keys'] if key["alias"] != alias]
        else:
            self.config['keys'] = []
        self.config['keys'].append({
            'alias': alias,
            'generation': generation,
            'file': file_path
        })
        return self

    def add_overlay(self,
                    overlay_class: str,
                    key_alias: str,
                    walkers: typing.List[WalkerDefinition],
                    bootstrappers: typing.List[BootstrapperDefinition],
                    initialize: typing.Dict[str, typing.Any],
                    on_start: typing.List[tuple],
                    allow_duplicate: bool = False) -> ConfigBuilder:
        """
        Add an overlay by its class name. You can choose from the default communities or register your own (see IPv8's
        ``extra_communities`` for the latter). Whatever key alias you choose for this overlay should be registered
        through add_key.

        The default communities include:

         - 'AttestationCommunity'
         - 'DiscoveryCommunity'
         - 'HiddenTunnelCommunity'
         - 'IdentityCommunity'
         - 'TunnelCommunity'
         - 'DHTDiscoveryCommunity'

        The initialize argument is a key-value mapping passed to the constructor (__init__) of the overlay.

        The on_start contains a list of tuples that specify the method to call on the overlay when IPv8 has initialized,
        coupled to the positional arguments to pass to the method, for example to call ``some_method(1, 2)``:

         [("some_method", 1, 2)]

        Lastly, IPv8 is capable of loading two distinct instances of the same overlay class. Set ``allow_duplicate``
        to explicitly allow this (usually this is programmer error).
        """
        if 'overlays' in self.config:
            if not allow_duplicate:
                self.config['overlays'] = [overlay for overlay in self.config['overlays']
                                           if overlay["class"] != overlay_class]
        else:
            self.config['overlays'] = []
        self.config['overlays'].append({
            'class': overlay_class,
            'key': key_alias,
            'walkers': [{
                'strategy': walker.strategy.value,
                'peers': walker.peers,
                'init': walker.init
            } for walker in walkers],
            'bootstrappers': [{
                'class': bootstrapper.bootstrapper.value,
                'init': bootstrapper.init
            } for bootstrapper in bootstrappers],
            'initialize': initialize,
            'on_start': on_start
        })
        return self


__all__ = ['Bootstrapper', 'BootstrapperDefinition', 'ConfigBuilder', 'DISPERSY_BOOTSTRAPPER', 'Strategy',
           'WalkerDefinition', 'default_bootstrap_defs', 'get_default_configuration']



import time

from ..base import TestBase
from ...dht.storage import Storage


class TestStorage(TestBase):

    def test_get_and_put(self):
        storage = Storage()

        storage.put(b'key', b'value1')
        self.assertEqual(storage.get(b'key'), [b'value1'])

        storage.put(b'key', b'value2')
        self.assertEqual(storage.get(b'key'), [b'value2', b'value1'])

        storage.put(b'key', b'value1')
        self.assertEqual(storage.get(b'key'), [b'value1', b'value2'])

    def test_items_older_than(self):
        storage = Storage()
        storage.put(b'key', b'value')
        storage.items[b'key'][0].last_update = time.time() - 1
        self.assertEqual(storage.items_older_than(0), [(b'key', b'value')])
        self.assertEqual(storage.items_older_than(10), [])

    def test_clean(self):
        storage = Storage()

        storage.put(b'key', b'value', max_age=60)
        storage.items[b'key'][0].last_update = time.time() - 120
        storage.clean()
        self.assertEqual(storage.get('key'), [])

        storage.put(b'key', b'value', 60)
        storage.items[b'key'][0].last_update = time.time()
        storage.clean()
        self.assertEqual(storage.get(b'key'), [b'value'])



import operator
import struct
from asyncio import Future, iscoroutine

maximum_integer = 2147483647

int2byte = struct.Struct(">B").pack
byte2int = operator.itemgetter(0)


def succeed(result):
    future = Future()
    future.set_result(result)
    return future


def fail(exception):
    future = Future()
    future.set_exception(exception)
    return future


def maybe_coroutine(func, *args, **kwargs):
    value = func(*args, **kwargs)
    if iscoroutine(value) or isinstance(value, Future):
        return value

    async def coro():
        return value
    return coro()


def coroutine(func):
    async def call_async(*args, **kwargs):
        return func(*args, **kwargs)
    return call_async


def strip_sha1_padding(s: bytes):
    return s[12:] if s.startswith(b'SHA-1\x00\x00\x00\x00\x00\x00\x00') else s



import logging
from binascii import hexlify

from . import DHTError
from ..messaging.anonymization.tunnel import IntroductionPoint, PEER_SOURCE_DHT
from ..messaging.lazy_payload import VariablePayload, vp_compile
from ..messaging.serialization import default_serializer
from ..peer import Peer


@vp_compile
class DHTIntroPointPayload(VariablePayload):
    names = ['address', 'last_seen', 'intro_pk', 'seeder_pk']
    format_list = ['ip_address', 'I', 'varlenH', 'varlenH']


class DHTCommunityProvider(object):
    """
    This class is a wrapper around the DHTCommunity and is used to discover introduction points
    for hidden services.
    """

    def __init__(self, dht_community, port):
        self.dht_community = dht_community
        self.port = port
        self.logger = logging.getLogger(self.__class__.__name__)

    async def peer_lookup(self, mid, peer=None):
        try:
            await self.dht_community.connect_peer(mid, peer)
        except DHTError as e:
            self.logger.debug("Failed to connect %s using the DHTCommunity (error: %s)", hexlify(mid), e)
            return

    async def lookup(self, info_hash):
        try:
            values = await self.dht_community.find_values(info_hash)
        except DHTError as e:
            self.logger.info("Failed to lookup %s on the DHTCommunity (error: %s)", hexlify(info_hash), e)
            return

        results = []
        for value, _ in values:
            try:
                payload, _ = default_serializer.unpack_serializable(DHTIntroPointPayload, value)
                intro_peer = Peer(b'LibNaCLPK:' + payload.intro_pk, payload.address)
                results.append(IntroductionPoint(intro_peer, b'LibNaCLPK:' + payload.seeder_pk,
                                                 PEER_SOURCE_DHT, payload.last_seen))
            except Exception as e:
                self.logger.info("Error during lookup %s on the DHTCommunity (error: %s)", hexlify(info_hash), e)
        self.logger.info("Looked up %s in the DHTCommunity, got %d results", hexlify(info_hash), len(results))
        return info_hash, results

    async def announce(self, info_hash, intro_point):
        # We strip away the LibNaCLPK part of the public key to avoid going over the DHT size limit.
        value = default_serializer.pack_serializable(DHTIntroPointPayload(intro_point.peer.address,
                                                                          intro_point.last_seen,
                                                                          intro_point.peer.public_key.key_to_bin()[10:],
                                                                          intro_point.seeder_pk[10:]))

        try:
            await self.dht_community.store_value(info_hash, value)
        except DHTError as e:
            self.logger.info("Failed to announce %s to the DHTCommunity (error: %s)", hexlify(info_hash), e)
        else:
            self.logger.info("Announced %s to the DHTCommunity", hexlify(info_hash))



from ..messaging.lazy_payload import VariablePayload, vp_compile
from ..messaging.serialization import Payload


def encode_connection_type(type):
    if type == u"public":
        return (1, 0)
    if type == u"symmetric-NAT":
        return (1, 1)
    return (0, 0)


def decode_connection_type(bit_0, bit_1):
    bits = (bit_0, bit_1)
    if bits == (0, 0):
        return u"unknown"
    if bits == (1, 0):
        return u"public"
    if bits == (1, 1):
        return u"symmetric-NAT"


class IntroductionRequestPayload(Payload):

    msg_id = 246
    format_list = ['ipv4', 'ipv4', 'ipv4', 'bits', 'H', 'raw']

    def __init__(self, destination_address, source_lan_address, source_wan_address, advice, connection_type,
                 identifier, extra_bytes, supports_new_style=True):
        """
        Create the payload for an introduction-request message.

        DESTINATION_ADDRESS is the address of the receiver.  Effectively this should be the
        wan address that others can use to contact the receiver.

        SOURCE_LAN_ADDRESS is the lan address of the sender.  Nodes in the same LAN
        should use this address to communicate.

        SOURCE_WAN_ADDRESS is the wan address of the sender.  Nodes not in the same
        LAN should use this address to communicate.

        ADVICE is a boolean value.  When True the receiver will introduce the sender to a new
        node.  This introduction will be facilitated by the receiver sending a puncture-request
        to the new node.

        CONNECTION_TYPE is a unicode string indicating the connection type that the message
        creator has.  Currently the following values are supported: u"unknown", u"public", and
        u"symmetric-NAT".

        IDENTIFIER is a number that must be given in the associated introduction-response.  This
        number allows to distinguish between multiple introduction-response messages.

        EXTRA_BYTES is a string that can be used to piggyback extra information.
        """
        super(IntroductionRequestPayload, self).__init__()
        self.destination_address = destination_address
        self.source_lan_address = source_lan_address
        self.source_wan_address = source_wan_address
        self.advice = advice
        self.supports_new_style = supports_new_style
        self.connection_type = connection_type
        self.identifier = identifier % 65536
        self.extra_bytes = extra_bytes

    def to_pack_list(self):
        encoded_connection_type = encode_connection_type(self.connection_type)
        data = [('ipv4', self.destination_address),
                ('ipv4', self.source_lan_address),
                ('ipv4', self.source_wan_address),
                ('bits', encoded_connection_type[0], encoded_connection_type[1], self.supports_new_style, 0, 0, 0, 0,
                 self.advice),
                ('H', self.identifier),
                ('raw', self.extra_bytes)]
        return data

    @classmethod
    def from_unpack_list(cls, destination_address, source_lan_address, source_wan_address,
                         connection_type_0, connection_type_1, supports_new_style, dflag1, dflag2, tunnel, sync, advice,
                         identifier, extra_bytes):
        args = [destination_address,
                source_lan_address,
                source_wan_address,
                [True, False][advice],
                decode_connection_type(connection_type_0, connection_type_1),
                identifier,
                extra_bytes,
                supports_new_style]

        return IntroductionRequestPayload(*args)


@vp_compile
class NewIntroductionRequestPayload(VariablePayload):

    msg_id = 234
    format_list = ['ip_address', 'ip_address', 'ip_address', 'H', 'bits', 'raw']
    names = ["destination_address", "source_lan_address", "source_wan_address", "identifier", "flag0", "flag1",
             "flag2", "flag3", "flag4", "flag5", "flag6", "flag7", "extra_bytes"]


class IntroductionResponsePayload(Payload):

    msg_id = 245
    format_list = ['ipv4', 'ipv4', 'ipv4', 'ipv4', 'ipv4', 'bits', 'H', 'raw']

    def __init__(self, destination_address, source_lan_address, source_wan_address, lan_introduction_address,
                 wan_introduction_address, connection_type, identifier, extra_bytes, supports_new_style=True,
                 intro_supports_new_style=False, peer_limit_reached=False):
        """
        Create the payload for an introduction-response message.

        DESTINATION_ADDRESS is the address of the receiver.  Effectively this should be the
        wan address that others can use to contact the receiver.

        SOURCE_LAN_ADDRESS is the lan address of the sender.  Nodes in the same LAN
        should use this address to communicate.

        SOURCE_WAN_ADDRESS is the wan address of the sender.  Nodes not in the same
        LAN should use this address to communicate.

        LAN_INTRODUCTION_ADDRESS is the lan address of the node that the sender
        advises the receiver to contact.  This address is zero when the associated request did
        not want advice.

        WAN_INTRODUCTION_ADDRESS is the wan address of the node that the sender
        advises the receiver to contact.  This address is zero when the associated request did
        not want advice.

        CONNECTION_TYPE is a unicode string indicating the connection type that the message
        creator has.  Currently the following values are supported: u"unknown", u"public", and
        u"symmetric-NAT".

        IDENTIFIER is a number that was given in the associated introduction-request.  This
        number allows to distinguish between multiple introduction-response messages.

        EXTRA_BYTES is a string that can be used to piggyback extra information.

        When the associated request wanted advice the sender will also sent a puncture-request
        message to either the lan_introduction_address or the wan_introduction_address
        (depending on their positions).  The introduced node must sent a puncture message to the
        receiver to punch a hole in its NAT.
        """
        super(IntroductionResponsePayload, self).__init__()
        self.destination_address = destination_address
        self.source_lan_address = source_lan_address
        self.source_wan_address = source_wan_address
        self.lan_introduction_address = lan_introduction_address
        self.wan_introduction_address = wan_introduction_address
        self.connection_type = connection_type
        self.supports_new_style = supports_new_style
        self.intro_supports_new_style = intro_supports_new_style
        self.peer_limit_reached = peer_limit_reached
        self.identifier = identifier % 65536
        self.extra_bytes = extra_bytes

    def to_pack_list(self):
        encoded_connection_type = encode_connection_type(self.connection_type)
        data = [('ipv4', self.destination_address),
                ('ipv4', self.source_lan_address),
                ('ipv4', self.source_wan_address),
                ('ipv4', self.lan_introduction_address),
                ('ipv4', self.wan_introduction_address),
                ('bits', encoded_connection_type[0], encoded_connection_type[1], 0, self.supports_new_style,
                 self.intro_supports_new_style, self.peer_limit_reached, 0, 0),
                ('H', self.identifier),
                ('raw', self.extra_bytes)]
        return data

    @classmethod
    def from_unpack_list(cls, destination_address, source_lan_address, source_wan_address,
                         introduction_lan_address, introduction_wan_address,
                         connection_type_0, connection_type_1, dflag0, supports_new_style, intro_supports_new_style,
                         peer_limit_reached, dflag4, dflag5, identifier, extra_bytes):
        args = [destination_address,
                source_lan_address,
                source_wan_address,
                introduction_lan_address,
                introduction_wan_address,
                decode_connection_type(connection_type_0, connection_type_1),
                identifier,
                extra_bytes,
                supports_new_style,
                intro_supports_new_style,
                peer_limit_reached]

        return IntroductionResponsePayload(*args)


@vp_compile
class NewIntroductionResponsePayload(VariablePayload):

    msg_id = 233
    format_list = ['ip_address', 'ip_address', 'ip_address', 'ip_address', 'ip_address', 'H', 'bits', 'raw']
    names = ["destination_address", "source_lan_address", "source_wan_address", "lan_introduction_address",
             "wan_introduction_address", "identifier", "intro_supports_new_style", "flag1", "flag2", "flag3",
             "flag4", "flag5", "flag6", "flag7", "extra_bytes"]


class PunctureRequestPayload(Payload):

    msg_id = 250
    format_list = ['ipv4', 'ipv4', 'H']

    def __init__(self, lan_walker_address, wan_walker_address, identifier):
        """
        Create the payload for a puncture-request payload.

        LAN_WALKER_ADDRESS is the lan address of the node that the sender wants us to
        contact.  This contact attempt should punch a hole in our NAT to allow the node to
        connect to us.

        WAN_WALKER_ADDRESS is the wan address of the node that the sender wants us to
        contact.  This contact attempt should punch a hole in our NAT to allow the node to
        connect to us.

        IDENTIFIER is a number that was given in the associated introduction-request.  This
        number allows to distinguish between multiple introduction-response messages.

        TODO add connection type
        TODO add tunnel bit
        """
        super(PunctureRequestPayload, self).__init__()
        self.lan_walker_address = lan_walker_address
        self.wan_walker_address = wan_walker_address
        self.identifier = identifier % 65536

    def to_pack_list(self):
        data = [('ipv4', self.lan_walker_address),
                ('ipv4', self.wan_walker_address),
                ('H', self.identifier)]

        return data

    @classmethod
    def from_unpack_list(cls, lan_walker_address, wan_walker_address, identifier):
        args = [lan_walker_address,
                wan_walker_address,
                identifier]

        return PunctureRequestPayload(*args)


@vp_compile
class NewPunctureRequestPayload(VariablePayload):

    msg_id = 232
    format_list = ['ip_address', 'ip_address', 'H']
    names = ["lan_walker_address", "wan_walker_address", "identifier"]


class PuncturePayload(Payload):

    msg_id = 249
    format_list = ['ipv4', 'ipv4', 'H']

    def __init__(self, source_lan_address, source_wan_address, identifier):
        """
        Create the payload for a puncture message

        SOURCE_LAN_ADDRESS is the lan address of the sender.  Nodes in the same LAN
        should use this address to communicate.

        SOURCE_WAN_ADDRESS is the wan address of the sender.  Nodes not in the same
        LAN should use this address to communicate.

        IDENTIFIER is a number that was given in the associated introduction-request.  This
        number allows to distinguish between multiple introduction-response messages.
        """
        super(PuncturePayload, self).__init__()
        self.source_lan_address = source_lan_address
        self.source_wan_address = source_wan_address
        self.identifier = identifier % 65536

    def to_pack_list(self):
        data = [('ipv4', self.source_lan_address),
                ('ipv4', self.source_wan_address),
                ('H', self.identifier)]

        return data

    @classmethod
    def from_unpack_list(cls, lan_walker_address, wan_walker_address, identifier):
        args = [lan_walker_address,
                wan_walker_address,
                identifier]

        return PuncturePayload(*args)


@vp_compile
class NewPuncturePayload(VariablePayload):

    msg_id = 231
    format_list = ['ip_address', 'ip_address', 'H']
    names = ["source_lan_address", "source_wan_address", "identifier"]



"""
Ported from "The Weil Pairing on Elliptic Curves and Its Cryptographic Applications" - Appendix D by Alex Edward Aftuk.
"""
from .value import FP2Value


def esum(mod, p, q):
    """
    Perform Elliptic Curve addition of points P and Q over Fp^2.
    """
    if p == "O" and q == "O":
        return "O"
    if p == "O":
        return q
    if q == "O":
        return p
    x1, y1 = p
    x2, y2 = q
    if x1 == x2 and y1 == FP2Value(mod, -1) * y2:
        return "O"
    if x1 == x2:
        l = ((FP2Value(mod, 3) * x1 * x1) // (FP2Value(mod, 2) * y1)).normalize()
    else:
        l = ((y1 - y2) // (x1 - x2)).normalize()
    x3 = l * l - x1 - x2
    y3 = l * (x3 - x1) + y1
    return x3.normalize(), (FP2Value(mod, -1) * y3).normalize()


def H(mod, p, q, x, y):
    """
    Perform the h_{T,T} function for the Miller calculation with divisors P and Q for coordinate (x,y).
    """
    x1, y1 = p
    x2, y2 = q
    if x1 == x2 and y1 == FP2Value(mod, -1) * y2:
        return (x - x1).normalize()
    if x1 == x2 and y1 == y2:
        l = (FP2Value(mod, 3) * x1 * x1) // (FP2Value(mod, 2) * y1)
        return ((y - y1 - l * (x - x1)) // (x + (x1 + x2) - l * l)).normalize()
    l = (y2 - y1) // (x2 - x1)
    return ((y - y1 - l * (x - x1)) // (x + (x1 + x2) - l * l)).normalize()


def millercalc(mod, M, p, R):
    """
    Perform the Miller calculation for message M point P and coordinates given by R.
    """
    mlist = list(reversed([int(c) for c in str(bin(M))[2:]]))
    T = p
    f = FP2Value(mod, 1)
    for i in reversed(list(range(len(mlist) - 1))):
        f = (f * f * H(mod, T, T, R[0], R[1])).normalize()
        T = esum(mod, T, T)
        if mlist[i] == 1:
            f = (f * H(mod, T, p, R[0], R[1])).normalize()
            T = esum(mod, T, p)
    return f


def weilpairing(mod, m, P, Q, S):
    """
    Create a Weil pairing for message m, points P and Q and DH secret S.
    """
    nS = [S[0], FP2Value(mod, -1) * S[1]]
    A = millercalc(mod, m, P, esum(mod, Q, S))
    B = millercalc(mod, m, P, S)
    C = millercalc(mod, m, Q, esum(mod, P, nS))
    D = millercalc(mod, m, Q, nS)
    wp = ((A * D) // (B * C))
    return wp.wp_nominator() * wp.wp_denom_inverse()



import abc
import hashlib


class IdentityAlgorithm(metaclass=abc.ABCMeta):

    def __init__(self, id_format, formats):
        self.id_format = id_format
        self.honesty_check = False

        if id_format not in formats:
            raise RuntimeError("Tried to initialize with illegal identity format")

    @abc.abstractmethod
    def generate_secret_key(self):
        """
        Generate a secret key.

        :return: the secret key
        """
        pass

    @abc.abstractmethod
    def load_secret_key(self, serialized):
        """
        Unserialize a secret key from the key material.

        :param serialized: the string of the private key
        :return: the private key
        """
        pass

    @abc.abstractmethod
    def load_public_key(self, serialized):
        """
        Unserialize a public key from the key material.

        :param serialized: the string of the public key
        :return: the public key
        """
        pass

    @abc.abstractmethod
    def get_attestation_class(self):
        """
        Return the Attestation (sub)class for serialization

        :return: the Attestation object
        :rtype: Attestation
        """
        pass

    @abc.abstractmethod
    def attest(self, PK, value):
        """
        Attest to a value for a certain public key.

        :param PK: the public key of the party we are attesting for
        :param value: the value we are attesting to
        :type value: str
        :return: the attestation string
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def certainty(self, value, aggregate):
        """
        The current certainty of the aggregate object representing a certain value.

        :param value: the value to match to
        :type value: str
        :param aggregate: the aggregate object
        :return: the matching factor [0.0-1.0]
        :rtype: float
        """
        pass

    @abc.abstractmethod
    def create_challenges(self, PK, attestation):
        """
        Create challenges for a certain counterparty.

        :param PK: the public key of the party we are challenging
        :type PK: BonehPublicKey
        :param attestation: the attestation information
        :type attestation: Attestation
        :return: the challenges to send
        :rtype: [str]
        """

    @abc.abstractmethod
    def create_challenge_response(self, SK, attestation, challenge):
        """
        Create an honest response to a challenge of our value.

        :param SK: our secret key
        :param attestation: the attestation information
        :type attestation: Attestation
        :param challenge: the challenge to respond to
        :return: the response to a challenge
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def create_certainty_aggregate(self, attestation):
        """
        Create an empty aggregate object, for matching to values.

        :param attestation: the attestation information
        :type attestation: Attestation
        :return: the aggregate object
        """
        pass

    @abc.abstractmethod
    def create_honesty_challenge(self, PK, value):
        """
        Use a known value to check for honesty.

        :param PK: the public key of the party we are challenging
        :param value: the value to use
        :type value: str
        :return: the challenge to send
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def process_honesty_challenge(self, value, response):
        """
        Given a response, check if it matches the expected value.

        :param value: the expected value
        :param response: the returned response
        :type response: str
        :return: if the value matches the response
        :rtype: bool
        """

    @abc.abstractmethod
    def process_challenge_response(self, aggregate, challenge, response):
        """
        Given a response, update the current aggregate.

        :param aggregate: the aggregate object
        :param challenge: the sent challenge
        :type challenge: str
        :param response: the response to introduce
        :type response: str
        :return: the new aggregate
        """
        pass


class Attestation(metaclass=abc.ABCMeta):
    """
    An attestation for a public key of a value.

    !!! Requires implementation of a `.id_format` field.
    """

    @abc.abstractmethod
    def serialize(self):
        """
        Serialize this Attestation to a string.

        :return: the serialized form of this attestation
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def serialize_private(self, PK):
        """
        Serialize this Attestation to a string, include shared secrets (not to be published!).

        :param PK: the public key to encode for
        :return: the serialized form of this attestation
        :rtype: str
        """
        pass

    @classmethod
    def unserialize(cls, s, id_format):
        """
        Given a string, create an Attestation object.

        :param s: the string to unserialize
        :type s: str
        :param id_format: the identity format
        :type id_format: str
        :return: the attestation object
        :rtype: Attestation
        """
        raise NotImplementedError()

    @classmethod
    def unserialize_private(cls, SK, s, id_format):
        """
        Given a string, create an Attestation object.
        The input contains shared secrets not to be published.

        :param SK: the secret key to decode with
        :param s: the string to unserialize
        :type s: str
        :param id_format: the identity format
        :type id_format: str
        :return: the attestation object
        :rtype: Attestation
        """
        raise NotImplementedError()

    def get_hash(self):
        """
        The hash over the public part of this Attestation.
        """
        return hashlib.sha1(self.serialize()).digest()


__all__ = ["IdentityAlgorithm", "Attestation"]



import asyncio
import base64
import json
import os
import typing

from .identity.community import IdentityCommunity, create_community
from .identity.manager import IdentityManager
from .wallet.community import AttestationCommunity
from ..keyvault.crypto import ECCrypto
from ..messaging.anonymization.hidden_services import HiddenTunnelCommunity
from ..types import IPv8, Peer
from ..util import succeed

AttributePointer = typing.Tuple[Peer, str]
MetadataDict = typing.Dict[str, str]


class CommunicationChannel(object):

    def __init__(self, attestation_community: AttestationCommunity, identity_community: IdentityCommunity):
        super(CommunicationChannel, self).__init__()
        self.attestation_overlay = attestation_community
        self.identity_overlay = identity_community

        self.attestation_requests: typing.Dict[AttributePointer, typing.Tuple[asyncio.Future, str]] = {}
        self.verify_requests: typing.Dict[AttributePointer, asyncio.Future] = {}
        self.verification_output: typing.Dict[bytes, typing.List[typing.Tuple[bytes, typing.Optional[float]]]] = {}
        self.attestation_metadata: typing.Dict[AttributePointer, MetadataDict] = {}

        self.attestation_overlay.set_attestation_request_callback(self.on_request_attestation)
        self.attestation_overlay.set_attestation_request_complete_callback(self.on_attestation_complete)
        self.attestation_overlay.set_verify_request_callback(self.on_verify_request)

    @property
    def public_key_bin(self) -> bytes:
        return self.identity_overlay.my_peer.public_key.key_to_bin()

    @property
    def peers(self) -> typing.List[Peer]:
        return self.identity_overlay.get_peers()

    @property
    def schemas(self) -> typing.List[str]:
        return list(self.attestation_overlay.schema_manager.formats.keys())

    def on_request_attestation(self, peer: Peer, attribute_name: str,
                               metadata: MetadataDict) -> asyncio.Future:
        """
        Return the measurement of an attribute for a certain peer.
        """
        future: asyncio.Future = asyncio.Future()
        self.attestation_requests[(peer, attribute_name)] = (future, json.dumps(metadata))
        self.attestation_metadata[(peer, attribute_name)] = metadata
        return future

    def on_attestation_complete(self, for_peer: Peer, attribute_name: str, attribute_hash: bytes, id_format: str,
                                from_peer: typing.Optional[Peer] = None) -> None:
        """
        Callback for when an attestation has been completed for another peer.
        We can now sign for it.
        """
        metadata = self.attestation_metadata.get((for_peer, attribute_name), None)
        if for_peer == self.identity_overlay.my_peer:
            if from_peer == self.identity_overlay.my_peer:
                self.identity_overlay.self_advertise(attribute_hash, attribute_name, id_format, metadata)
            else:
                self.identity_overlay.request_attestation_advertisement(from_peer,  # type:ignore
                                                                        attribute_hash, attribute_name,
                                                                        id_format, metadata)
        else:
            self.identity_overlay.add_known_hash(attribute_hash, attribute_name, for_peer.public_key.key_to_bin(),
                                                 metadata)

    def on_verify_request(self, peer: Peer, attribute_hash: bytes) -> asyncio.Future:
        """
        Return the measurement of an attribute for a certain peer.
        """
        metadata = self.identity_overlay.get_attestation_by_hash(attribute_hash)
        if not metadata:
            return succeed(None)
        attribute_name = json.loads(metadata.serialized_json_dict)["name"]
        future: asyncio.Future = asyncio.Future()
        self.verify_requests[(peer, attribute_name)] = future
        return future

    def on_verification_results(self, attribute_hash: bytes, values: typing.List[float]) -> None:
        """
        Callback for when verification has concluded.
        """
        references = self.verification_output[attribute_hash]
        out = []
        for i in range(len(references)):
            out.append((references[i][0], values[i]))
        self.verification_output[attribute_hash] = out  # type:ignore

    def _drop_identity_table_data(self) -> typing.List[bytes]:
        """
        Remove all metadata from the identity community.

        :return: the list of attestation hashes which have been removed
        :rtype: [database_blob]
        """
        database = self.identity_overlay.identity_manager.database
        attestation_hashes = [t.content_hash for t in database.get_tokens_for(self.identity_overlay.my_peer.public_key)]

        database.executescript("BEGIN TRANSACTION; "
                               "DELETE FROM Tokens WHERE public_key = ?; "
                               "DELETE FROM Metadata WHERE public_key = ?; "
                               "DELETE FROM Attestations WHERE public_key = ?; "
                               "COMMIT;",
                               (self.public_key_bin, self.public_key_bin, self.public_key_bin))
        database.commit()

        return attestation_hashes

    def _drop_attestation_table_data(self, attestation_hashes: typing.List[bytes]) -> None:
        """
        Remove all attestation data (claim based keys and ZKP blobs) by list of attestation hashes.

        :param attestation_hashes: hashes to remove
        :type attestation_hashes: [database_blob]
        :returns: None
        """
        if not attestation_hashes:
            return

        self.attestation_overlay.database.execute(("DELETE FROM %s" % self.attestation_overlay.database.db_name)
                                                  + " WHERE hash IN ("
                                                  + ", ".join(c for c in "?" * len(attestation_hashes))
                                                  + ")",
                                                  attestation_hashes)
        self.attestation_overlay.database.commit()

    def remove(self) -> None:
        """
        Remove this pseudonym from existence.
        """
        self._drop_attestation_table_data(self._drop_identity_table_data())
        self.attestation_requests.clear()

    def get_my_attributes(self) -> typing.Dict[bytes, typing.Tuple[str,
                                                                   MetadataDict,
                                                                   typing.List[bytes]]]:
        return self.get_attributes(self.identity_overlay.my_peer)

    def get_attributes(self, peer: Peer) -> typing.Dict[bytes,
                                                        typing.Tuple[str,
                                                                     MetadataDict,
                                                                     typing.List[bytes]]]:
        pseudonym = self.identity_overlay.identity_manager.get_pseudonym(peer.public_key)
        out = {}
        for credential in pseudonym.get_credentials():
            attestations = list(credential.attestations)
            attesters = []
            for attestation in attestations:
                attesters.append(self.identity_overlay.identity_manager.database.get_authority(attestation))
            attribute_hash = pseudonym.tree.elements[credential.metadata.token_pointer].content_hash
            json_metadata = json.loads(credential.metadata.serialized_json_dict)
            out[attribute_hash] = (json_metadata["name"], json_metadata, attesters)
        return out

    def request_attestation(self, peer: Peer, attribute_name: str, id_format: str,
                            metadata: MetadataDict) -> None:
        key = self.attestation_overlay.get_id_algorithm(id_format).generate_secret_key()
        metadata.update({"id_format": id_format})
        self.attestation_metadata[(self.identity_overlay.my_peer, attribute_name)] = metadata
        self.attestation_overlay.request_attestation(peer, attribute_name, key, metadata)

    def attest(self, peer: Peer, attribute_name: str, value: bytes) -> None:
        outstanding = self.attestation_requests.pop((peer, attribute_name))
        outstanding[0].set_result(value)

    def import_blob(self, attribute_name: str, id_format: str, metadata: MetadataDict, value: bytes) -> None:
        metadata.update({"id_format": id_format})
        self.attestation_overlay.dump_blob(attribute_name, id_format, value, metadata)

    def allow_verification(self, peer: Peer, attribute_name: str) -> None:
        outstanding = self.verify_requests.pop((peer, attribute_name))
        outstanding.set_result(True)

    def disallow_verification(self, peer: Peer, attribute_name: str) -> None:
        outstanding = self.verify_requests.pop((peer, attribute_name))
        outstanding.set_result(False)

    def verify(self, peer: Peer, attribute_hash: bytes, reference_values: typing.List[bytes], id_format: str) -> None:
        self.verification_output[attribute_hash] = [(v, None) for v in reference_values]
        self.attestation_overlay.verify_attestation_values(peer.address, attribute_hash, reference_values,
                                                           self.on_verification_results, id_format)


class CommunicationManager(object):

    def __init__(self, ipv8_instance: IPv8, pseudonym_folder: str = "pseudonyms", working_directory: str = None):
        super(CommunicationManager, self).__init__()

        self.ipv8_instance = ipv8_instance
        self.channels: typing.Dict[bytes, CommunicationChannel] = {}

        self.name_to_channel: typing.Dict[str, CommunicationChannel] = {}

        self.crypto = ECCrypto()

        loaded_community = ipv8_instance.get_overlay(IdentityCommunity)
        self.identity_manager = None if loaded_community is None else loaded_community.identity_manager

        if working_directory is None:
            working_directory = ipv8_instance.configuration.get("working_directory", ".")

        self.working_directory = working_directory
        self.pseudonym_folder = os.path.join(self.working_directory, pseudonym_folder)

    def lazy_identity_manager(self) -> IdentityManager:
        """
        Lazy load the IdentityManager.
        """
        if self.identity_manager is None:
            self.identity_manager = IdentityManager(self.working_directory if self.working_directory == ":memory:"
                                                    else os.path.join(self.working_directory, "sqlite", "identity.db"))
        return self.identity_manager

    async def load(self, name: str, rendezvous_token: typing.Optional[str] = None) -> CommunicationChannel:
        """
        Load a pseudonym.
        """
        if name in self.name_to_channel:
            return self.name_to_channel[name]

        os.makedirs(self.pseudonym_folder, exist_ok=True)
        pseudonym_file = os.path.join(self.pseudonym_folder, name)
        if os.path.exists(pseudonym_file):
            with open(pseudonym_file, 'rb') as file_handle:
                private_key = self.crypto.key_from_private_bin(file_handle.read())
        else:
            private_key = self.crypto.generate_key("curve25519")
            with open(pseudonym_file, 'wb') as file_handle:
                file_handle.write(private_key.key_to_bin())

        public_key = private_key.pub().key_to_bin()
        if public_key not in self.channels:
            tunnel_community = self.ipv8_instance.get_overlay(HiddenTunnelCommunity)
            decoded_rendezvous_token = (base64.b64decode(rendezvous_token.encode())
                                        if rendezvous_token is not None else None)
            identity_overlay = await create_community(private_key, self.ipv8_instance, self.lazy_identity_manager(),
                                                      working_directory=self.working_directory,
                                                      anonymize=tunnel_community is not None,
                                                      rendezvous_token=decoded_rendezvous_token)
            attestation_overlay = AttestationCommunity(identity_overlay.my_peer, identity_overlay.endpoint,
                                                       identity_overlay.network,
                                                       working_directory=self.working_directory,
                                                       anonymize=tunnel_community is not None)
            identity_overlay.endpoint.set_tunnel_community(tunnel_community)
            self.channels[public_key] = CommunicationChannel(attestation_overlay, identity_overlay)
            self.name_to_channel[name] = self.channels[public_key]

        return self.name_to_channel[name]

    async def unload(self, name: str) -> None:
        """
        Unload a pseudonym.
        """
        if name in self.name_to_channel:
            communication_channel = self.name_to_channel.pop(name)
            self.channels.pop(communication_channel.public_key_bin)
            await self.ipv8_instance.unload_overlay(communication_channel.identity_overlay)
            await communication_channel.attestation_overlay.unload()
            communication_channel.identity_overlay.endpoint.close()

    async def remove(self, name: str) -> None:
        """
        Remove a pseudonym from existence by name.
        """
        if name in self.name_to_channel:
            self.name_to_channel.pop(name).remove()
            await self.unload(name)
            pseudonym_file = os.path.join(self.pseudonym_folder, name)
            if os.path.exists(pseudonym_file):
                os.remove(pseudonym_file)

    async def shutdown(self) -> None:
        for name in list(self.name_to_channel):
            await self.unload(name)

    def list_names(self) -> typing.List[str]:
        """
        List all known pseudonyms.
        """
        if not os.path.exists(self.pseudonym_folder):
            return []
        return os.listdir(self.pseudonym_folder)

    def list_loaded(self) -> typing.List[str]:
        """
        List all loaded pseudonyms by name.
        """
        return [name for name in self.list_names() if name in self.name_to_channel]



import binascii
import json
import os

from .gabi.attributes import make_attribute_list
from .gabi.keys import DefaultSystemParameters
from .gabi.proofs import createChallenge
from .wrappers import challenge_response, serialize_proof_d, unserialize_proof_d
from ..primitives.structs import ipack, iunpack
from ...identity_formats import Attestation, IdentityAlgorithm


class IRMAAttestation(Attestation):

    def __init__(self, sign_date, proofd, z=None):
        self.sign_date = sign_date
        self.proofd = proofd
        self.z = z

    def serialize(self):
        return ipack(self.sign_date) + serialize_proof_d(self.proofd)

    def serialize_private(self, PK):
        return ipack(self.z) + ipack(self.sign_date) + serialize_proof_d(self.proofd)

    @classmethod
    def unserialize(cls, s, id_format):
        sign_date, rem = iunpack(s)
        return IRMAAttestation(sign_date, unserialize_proof_d(rem))

    @classmethod
    def unserialize_private(cls, SK, s, id_format):
        z, rem = iunpack(s)
        sign_date, rem = iunpack(rem)
        return IRMAAttestation(sign_date, unserialize_proof_d(rem), z)


class KeyStub(object):

    def public_key(self):
        return self

    def serialize(self):
        return b''

    @classmethod
    def unserialize(cls, s):
        return KeyStub()


class IRMAExactAlgorithm(IdentityAlgorithm):

    def __init__(self, id_format, formats):
        super(IRMAExactAlgorithm, self).__init__(id_format, formats)

        # Check algorithm match
        if formats[id_format]["algorithm"] != "irmaexact":
            raise RuntimeError("Identity format linked to wrong algorithm")

        self.issuer_pk = formats[self.id_format]["issuer_pk"]
        self.attribute_order = formats[self.id_format]["order"]
        self.validity = formats[self.id_format]["validity"]

        self.base_meta = {
            u"credential": formats[self.id_format]["credential"],
            u"keyCounter": formats[self.id_format]["keyCounter"],
            u"validity": formats[self.id_format]["validity"]
        }

        self.system_parameters = DefaultSystemParameters[1024]
        self.challenge_count = 8

    def generate_secret_key(self):
        return KeyStub()

    def load_secret_key(self, serialized):
        return KeyStub()

    def load_public_key(self, serialized):
        return KeyStub()

    def get_attestation_class(self):
        return IRMAAttestation

    def attest(self, PK, value):
        raise NotImplementedError("Only import_blob is supported (now) for IRMA.")

    def certainty(self, value, aggregate):
        value_json = {u"attributes": json.loads(value)}
        value_json.update(self.base_meta)
        attestation = aggregate['attestation']
        attr_ints, sign_date = make_attribute_list(value_json, self.attribute_order,
                                                   (self.validity, attestation.sign_date))
        reconstructed_attr_map = {}
        for i in range(len(attr_ints)):
            reconstructed_attr_map[i + 1] = attr_ints[i]

        verified = 0.0
        failure = False
        for k, v in aggregate.items():
            if k != 'attestation' and v:
                challenge_verif, _ = iunpack(k)
                p = attestation.proofd.Copy()
                p.ADisclosed = reconstructed_attr_map
                Ap, Zp = p.ChallengeContribution(self.issuer_pk)
                p.C, _ = iunpack(v)
                reconstructed_challenge = createChallenge(challenge_verif, challenge_verif, [Ap, Zp], False)
                if p.VerifyWithChallenge(self.issuer_pk, reconstructed_challenge):
                    verified += 1.0
                else:
                    failure = True

        return 0.0 if failure else (verified / self.challenge_count)

    def create_challenges(self, PK, attestation):
        return [ipack(int(binascii.hexlify(os.urandom(32)), 16) % self.issuer_pk.N)
                for _ in range(self.challenge_count)]

    def create_challenge_response(self, SK, attestation, challenge):
        return challenge_response(attestation.proofd, attestation.z, challenge)

    def create_certainty_aggregate(self, attestation):
        return {'attestation': attestation}

    def create_honesty_challenge(self, PK, value):
        raise NotImplementedError()

    def process_honesty_challenge(self, value, response):
        raise NotImplementedError()

    def process_challenge_response(self, aggregate, challenge, response):
        aggregate[challenge] = response

    def import_blob(self, blob):
        blob_json = json.loads(blob)

        sign_date = blob_json["sign_date"]
        proofd = unserialize_proof_d(binascii.unhexlify(blob_json["proofd"]))
        z = blob_json["z"]

        inst = self.get_attestation_class()(sign_date, proofd, z)

        return inst.serialize_private(None), None



from .endpoint import AutoMockEndpoint
from ...keyvault.crypto import default_eccrypto
from ...peer import Peer
from ...peerdiscovery.community import DiscoveryCommunity
from ...peerdiscovery.network import Network


class MockCommunity(DiscoveryCommunity):

    def __init__(self):
        endpoint = AutoMockEndpoint()
        endpoint.open()
        network = Network()
        peer = Peer(default_eccrypto.generate_key(u"very-low"), endpoint.wan_address)
        super(MockCommunity, self).__init__(peer, endpoint, network)
        # workaround for race conditions in deliver_messages
        self._use_main_thread = False
        self.my_estimated_lan = endpoint.lan_address
        self.my_estimated_wan = endpoint.wan_address

    def bootstrap(self):
        super(MockCommunity, self).bootstrap()
        self.last_bootstrap = 0



from ..mocking.endpoint import AutoMockEndpoint
from ...messaging.anonymization.tunnel import DataChecker, TunnelExitSocket
from ...messaging.interfaces.endpoint import EndpointListener


class MockTunnelExitSocket(TunnelExitSocket, EndpointListener):

    def __init__(self, parent):
        self.endpoint = AutoMockEndpoint()
        self.endpoint.open()
        self.parent = parent

        TunnelExitSocket.__init__(self, parent.circuit_id, parent.peer, parent.overlay)
        EndpointListener.__init__(self, self.endpoint, main_thread=False)

        self.endpoint.add_listener(self)

    def enable(self):
        self.enabled = True

    def sendto(self, data, destination):
        if DataChecker.could_be_bt(data) or DataChecker.could_be_ipv8(data):
            self.endpoint.send(destination, data)
        else:
            raise AssertionError("Attempted to exit data which is not allowed: %s" % repr(data))

    def on_packet(self, packet):
        source_address, data = packet
        self.datagram_received(data, source_address)

    async def close(self):
        await self.shutdown_task_manager()
        await self.parent.close()



"""
Copyright (c) 2016, Maarten Everts
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/gabi
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""
from cryptography.hazmat.primitives.asymmetric.rsa import _modinv  # type:ignore

from pyasn1.codec.ber.encoder import encode
from pyasn1.type import univ

from ...primitives.attestation import sha256_as_int
from ...primitives.value import FP2Value


class Record(univ.SequenceOf):
    componentType = univ.Integer()


def custom_asn1_marshal(values):
    return encode(values, asn1Spec=Record())


def hashCommit(values, issig):
    tmp = [True] if issig else []
    tmp = tmp + [len(values)] + values
    r = custom_asn1_marshal(tmp)
    if issig:
        indx = r.find(b'\x02\x01\x01')
        r = r[:indx] + b'\x01\x01\xFF' + r[indx + 3:]
    return sha256_as_int(r)


def createChallenge(context, nonce, contributions, issig):
    return hashCommit([context] + contributions + [nonce], issig)


class ProofU(object):

    def __init__(self, U, C, VPrimeResponse, SResponse):
        self.U = U
        self.C = C
        self.VPrimeResponse = VPrimeResponse
        self.SResponse = SResponse

    def MergeProofP(self, proofP, pk):
        self.U = (self.U * proofP.P) % pk.N
        self.SResponse = self.SResponse + proofP.SResponse

    def Verify(self, pk, context, nonce):
        return self.VerifyWithChallenge(pk, createChallenge(context, nonce, self.ChallengeContribution(pk), False))

    def correctResponseSizes(self, pk):
        maximum = (1 << (pk.Params.LvPrimeCommit + 1)) - 1
        minimum = -maximum
        return self.VPrimeResponse >= minimum and self.VPrimeResponse <= maximum

    def VerifyWithChallenge(self, pk, reconstructedChallenge):
        return self.correctResponseSizes(pk) and self.C == reconstructedChallenge

    def reconstructUcommit(self, pk):
        Uc = FP2Value(pk.N, self.U).intpow(-self.C)
        Sv = FP2Value(pk.N, pk.S).intpow(self.VPrimeResponse)
        R0s = FP2Value(pk.N, pk.R[0]).intpow(self.SResponse)
        return (Uc * Sv * R0s).a

    def SecretKeyResponse(self):
        return self.SResponse

    def Challenge(self):
        return self.C

    def ChallengeContribution(self, pk):
        return [self.U, self.reconstructUcommit(pk)]


class ProofS(object):

    def __init__(self, C, EResponse):
        self.C = C
        self.EResponse = EResponse

    def Verify(self, pk, signature, context, nonce):
        exponent = self.EResponse * signature.E + self.C
        ACommit = FP2Value(pk.N, signature.A).intpow(exponent).a
        Q = FP2Value(pk.N, signature.A).intpow(signature.E).a
        cPrime = hashCommit([context, Q, signature.A, nonce, ACommit], False)
        return self.C == cPrime


class ProofD(object):

    def __init__(self, C, A, EResponse, VResponse, AResponses, ADisclosed):
        self.C = C
        self.A = A
        self.EResponse = EResponse
        self.VResponse = VResponse
        self.AResponses = AResponses
        self.ADisclosed = ADisclosed

    def MergeProofP(self, proofP, pk):
        self.AResponses[0] += proofP.SResponse

    def correctResponseSizes(self, pk):
        maximum = (1 << (pk.Params.LmCommit + 1)) - 1
        minimum = -maximum

        for aResponse in self.AResponses:
            if aResponse < minimum or aResponse > maximum:
                return False

        maximum = (1 << (pk.Params.LeCommit + 1)) - 1
        minimum = -maximum

        if self.EResponse < minimum or self.EResponse > maximum:
            return False

        return True

    def reconstructZ(self, pk):
        numerator = 1 << (pk.Params.Le - 1)
        numerator = FP2Value(pk.N, self.A).intpow(numerator).a

        for i, exp in self.ADisclosed.items():
            if exp.bit_length() > pk.Params.Lm:
                exp = sha256_as_int(str(exp))
            numerator *= FP2Value(pk.N, pk.R[i]).intpow(exp).a
        numerator = numerator % pk.N

        known = pk.Z * _modinv(numerator, pk.N)
        knownC = FP2Value(pk.N, known).intpow(-self.C).a
        Ae = FP2Value(pk.N, self.A).intpow(self.EResponse).a
        Sv = FP2Value(pk.N, pk.S).intpow(self.VResponse).a
        Rs = 1
        for i, response in self.AResponses.items():
            Rs *= FP2Value(pk.N, pk.R[i]).intpow(response).a
        Z = (knownC * Ae * Rs * Sv) % pk.N
        return Z

    def Verify(self, pk, context, nonce1, issig):
        return self.VerifyWithChallenge(pk, createChallenge(context, nonce1, self.ChallengeContribution(pk), issig))

    def VerifyWithChallenge(self, pk, reconstructedChallenge):
        return self.correctResponseSizes(pk) and self.C == reconstructedChallenge

    def ChallengeContribution(self, pk):
        return [self.A, self.reconstructZ(pk)]

    def SecretKeyResponse(self):
        return self.AResponses[0]

    def Challenge(self):
        return self.C

    def Copy(self):
        ADisclosed = {}
        for k, v in self.ADisclosed.items():
            ADisclosed[k] = v
        return ProofD(self.C, self.A, self.EResponse, self.VResponse, self.AResponses, ADisclosed)


class ProofP(object):

    def __init__(self, P, C, SResponse):
        self.P = P
        self.C = C
        self.SResponse = SResponse


class ProofPCommitment(object):

    def __init__(self, P, Pcommit):
        self.P = P
        self.Pcommit = Pcommit



from .asyncio_endpoint import AsyncioEndpoint
from .attestation_endpoint import AttestationEndpoint
from .base_endpoint import BaseEndpoint
from .dht_endpoint import DHTEndpoint
from .identity_endpoint import IdentityEndpoint
from .isolation_endpoint import IsolationEndpoint
from .network_endpoint import NetworkEndpoint
from .noblock_dht_endpoint import NoBlockDHTEndpoint
from .overlays_endpoint import OverlaysEndpoint
from .tunnel_endpoint import TunnelEndpoint


class RootEndpoint(BaseEndpoint):
    """
    The root endpoint of the HTTP API is the root resource in the request tree.
    It will dispatch requests regarding torrents, channels, settings etc to the right child endpoint.
    """

    def setup_routes(self):
        endpoints = {'/asyncio': AsyncioEndpoint,
                     '/attestation': AttestationEndpoint,
                     '/dht': DHTEndpoint,
                     '/identity': IdentityEndpoint,
                     '/isolation': IsolationEndpoint,
                     '/network': NetworkEndpoint,
                     '/noblockdht': NoBlockDHTEndpoint,
                     '/overlays': OverlaysEndpoint,
                     '/tunnel': TunnelEndpoint}
        for path, ep_cls in endpoints.items():
            self.add_endpoint(path, ep_cls())
